from typing import Optional, List

from pydantic import BaseModel, Field


class BaseResponseWarning(BaseModel):
    colorHex: str = Field(None, alias="colorHex")
    message: Optional[str] = Field(None, alias='message')

class TheStageBaseResponse(BaseModel):
    message: Optional[str] = Field(None, alias='message')
    isSuccess: Optional[bool] = Field(None, alias='isSuccess')
    fieldsWithError: Optional[List[str]] = Field(None, alias='fieldsWithError')
    warningsImmediate: Optional[List[BaseResponseWarning]] = Field(None, alias='warningsImmediate')
    warningsOnExit: Optional[List[BaseResponseWarning]] = Field(None, alias='warningsOnExit')


# TODO: add Generic then back doing that
class TheStageBasePaginatedResponse(TheStageBaseResponse):
    fieldsWithError: List[str] = Field(None, alias='fieldsWithError')



from typing import Optional

import click
import typer
from git import Commit
from rich import print

from thestage.color_scheme.color_scheme import ColorScheme
from thestage.config.business.config_provider import ConfigProvider
from thestage.docker_container.communication.docker_container_api_client import DockerContainerApiClient
from thestage.docker_container.dto.container_response import DockerContainerDto
from thestage.git.communication.git_client import GitLocalClient
from thestage.global_dto.enums.yes_no_response import YesOrNoResponse
from thestage.helpers.error_handler import error_handler
from thestage.i18n.translation import __
from thestage.project.business.project_service import ProjectService
from thestage.project.dto.project_config import ProjectConfig
from thestage.services.abstract_service import AbstractService
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList
from thestage.services.filesystem_service import FileSystemService
from thestage.task.business.mapper.task_mapper import TaskMapper
from thestage.task.communication.task_api_client import TaskApiClient
from thestage.task.dto.run_task_response import RunTaskResponse
from thestage.task.dto.task import Task
from thestage.task.dto.task_entity import TaskEntity


class TaskService(AbstractService):
    def __init__(
            self,
            docker_container_api_client: DockerContainerApiClient,
            task_api_client: TaskApiClient,
            config_provider: ConfigProvider,
            git_local_client: GitLocalClient,
            file_system_service: FileSystemService,
            project_service: ProjectService,
    ):
        self.__docker_container_api_client = docker_container_api_client
        self.__task_api_client = task_api_client
        self.__config_provider = config_provider
        self.__git_local_client = git_local_client
        self.__file_system_service = file_system_service
        self.__project_service = project_service


    @error_handler()
    def project_run_task(
            self,
            run_command: str,
            docker_container_slug: str,
            docker_container_public_id: str,
            task_title: Optional[str] = None,
            commit_hash: Optional[str] = None,
            files_to_add: Optional[str] = None,
            is_skip_auto_commit: Optional[bool] = False,
    ) -> Optional[Task]:
        config = self.__config_provider.get_config()
        project_config: ProjectConfig = self.__project_service.get_fixed_project_config(do_check_default_container=True)
        if not project_config:
            typer.echo(__("No project found at the path: %path%. Initialize or clone a project first.",
                          {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        if not docker_container_public_id and not docker_container_slug and not project_config.default_container_public_id:
            typer.echo(__('Docker container ID or name is required'))
            raise typer.Exit(1)

        final_container_public_id = docker_container_public_id
        final_container_slug = docker_container_slug
        if not final_container_public_id and not final_container_slug:
            final_container_public_id = project_config.default_container_public_id
            typer.echo(
                f"Using default docker container for this project: '{project_config.default_container_public_id}'")

        container: DockerContainerDto = self.__docker_container_api_client.get_container(
            container_slug=final_container_slug,
            container_public_id=final_container_public_id
        )

        if container is None or container.is_deleted_permanently:
            if final_container_slug:
                typer.echo(f"Could not find container with name '{final_container_slug}'")
            if final_container_public_id:
                typer.echo(f"Could not find container with ID '{final_container_public_id}'")
            if project_config.default_container_public_id == final_container_public_id:
                project_config.default_container_public_id = None
                project_config.prompt_for_default_container = True
                self.__config_provider.save_project_config(project_config=project_config)
                typer.echo(f"Default container settings were reset")
            raise typer.Exit(1)

        if container.project.public_id != project_config.public_id:
            typer.echo(
                f"Provided container '{container.public_id}' is not related to project '{project_config.public_id}'")
            raise typer.Exit(1)

        if (project_config.prompt_for_default_container is None or project_config.prompt_for_default_container) and (
                docker_container_slug or docker_container_public_id) and (
                project_config.default_container_public_id != container.public_id):

            container_display_name = docker_container_slug or docker_container_public_id

            set_default_container_answer: str = typer.prompt(
                text=f"Would you like to set '{container_display_name}' as a default container for this project installation?",
                show_choices=True,
                default=YesOrNoResponse.YES.value,
                type=click.Choice([r.value for r in YesOrNoResponse]),
                show_default=True,
            )
            project_config.prompt_for_default_container = False
            if set_default_container_answer == YesOrNoResponse.YES.value:
                project_config.default_container_public_id = container.public_id

            self.__config_provider.save_project_config(project_config=project_config)

        has_wrong_args = files_to_add and commit_hash or is_skip_auto_commit and commit_hash or files_to_add and is_skip_auto_commit

        if has_wrong_args:
            warning_msg = f"[{ColorScheme.WARNING.value}][WARNING] You can provide only one of the following arguments: --commit-hash, --files-add, --skip-autocommit[{ColorScheme.WARNING.value}]"
            print(warning_msg)
            raise typer.Exit(1)

        if not is_skip_auto_commit and not commit_hash:
            is_git_folder = self.__git_local_client.is_present_local_git(path=config.runtime.working_directory)
            if not is_git_folder:
                typer.echo("Error: Working directory is not a git repository")
                raise typer.Exit(1)

            is_commit_allowed: bool = True
            has_changes = self.__git_local_client.has_changes_with_untracked(
                path=config.runtime.working_directory,
            )

            if self.__git_local_client.is_head_detached(path=config.runtime.working_directory):
                is_commit_allowed = False
                print(f"[{ColorScheme.GIT_HEADLESS.value}]HEAD is detached[{ColorScheme.GIT_HEADLESS.value}]")

                is_headless_commits_present = self.__git_local_client.is_head_committed_in_headless_state(
                    path=config.runtime.working_directory)
                if is_headless_commits_present:
                    print(
                        f"[{ColorScheme.GIT_HEADLESS.value}]Current commit created in detached HEAD state. Cannot use it to run the task. Consider using 'project checkout' command to return to a valid reference.[{ColorScheme.GIT_HEADLESS.value}]")
                    raise typer.Exit(1)

                if has_changes:
                    print(
                        f"[{ColorScheme.GIT_HEADLESS.value}]Local changes detected in detached head state. They will not impact the task execution.[{ColorScheme.GIT_HEADLESS.value}]")
                    response: YesOrNoResponse = typer.prompt(
                        text=__('Continue?'),
                        show_choices=True,
                        default=YesOrNoResponse.YES.value,
                        type=click.Choice([r.value for r in YesOrNoResponse]),
                        show_default=True,
                    )
                    if response == YesOrNoResponse.NO:
                        raise typer.Exit(0)

            if is_commit_allowed:
                if not self.__git_local_client.add_files_with_size_limit_or_warn(config.runtime.working_directory,
                                                                                 files_to_add):
                    warning_msg = f"[{ColorScheme.WARNING.value}][WARNING] Task was not started [{ColorScheme.WARNING.value}]"
                    print(warning_msg)
                    raise typer.Exit(1)

                diff_stat = self.__git_local_client.git_diff_stat(repo_path=config.runtime.working_directory)

                if has_changes and diff_stat:
                    branch_name = self.__git_local_client.get_active_branch_name(config.runtime.working_directory)

                    typer.echo(__('Active branch [%branch_name%] has uncommitted changes: %diff_stat_bottomline%', {
                        'diff_stat_bottomline': diff_stat,
                        'branch_name': branch_name,
                    }))

                    response: str = typer.prompt(
                        text=__('Commit changes?'),
                        show_choices=True,
                        default=YesOrNoResponse.YES.value,
                        type=click.Choice([r.value for r in YesOrNoResponse]),
                        show_default=True,
                    )
                    if response == YesOrNoResponse.NO.value:
                        typer.echo("Cannot run task with uncommitted changes - aborting")
                        raise typer.Exit(0)

                    commit_name = typer.prompt(
                        text=__('Provide commit message'),
                        show_choices=False,
                        type=str,
                        show_default=False,
                    )

                    if commit_name:
                        commit_result = self.__git_local_client.commit_local_changes(
                            path=config.runtime.working_directory,
                            name=commit_name
                        )

                        if commit_result:
                            # in docs not Commit object, on real - str
                            if isinstance(commit_result, str):
                                typer.echo(commit_result)
                    else:
                        typer.echo(__('Commit message cannot be empty'))
                        raise typer.Exit(0)
                else:
                    pass
                    # possible to push new empty branch - only that there's a wrong place to do so

                self.__git_local_client.push_changes(
                    path=config.runtime.working_directory,
                    deploy_key_path=project_config.deploy_key_path
                )
                typer.echo(__("Pushed changes to remote repository"))

        if not commit_hash:
            commit = self.__git_local_client.get_current_commit(path=config.runtime.working_directory)
            if not commit or not isinstance(commit, Commit):
                print('[red]Error: No current commit found in the local repository[/red]')
                raise typer.Exit(0)
            commit_hash = commit.hexsha
        else:
            commit = self.__git_local_client.get_commit_by_hash(path=config.runtime.working_directory,
                                                                commit_hash=commit_hash)
            if not commit or not isinstance(commit, Commit):
                print(f'[red]Error: commit \'{commit_hash}\' was not found in the local repository[/red]')
                raise typer.Exit(0)

        if not task_title:
            task_title = commit.message.strip() if commit.message else f'Task_{commit_hash}'
            if not commit.message:
                typer.echo(f'Commit message is empty. Task title is set to "{task_title}"')

        run_task_response: RunTaskResponse = self.__task_api_client.execute_project_task(
            project_public_id=project_config.public_id,
            docker_container_public_id=container.public_id,
            run_command=run_command,
            commit_hash=commit_hash,
            task_title=task_title,
        )
        if run_task_response:
            if run_task_response.message:
                print(f"[{ColorScheme.WARNING.value}]{run_task_response.message}[{ColorScheme.WARNING.value}]")
            if run_task_response.isSuccess and run_task_response.task:
                typer.echo(f"Task '{run_task_response.task.title}' has been scheduled successfully. Task ID: {run_task_response.task.public_id}")
                if run_task_response.tasksInQueue:
                    typer.echo(f"There are tasks in queue ahead of this new task:")
                    for queued_task_item in run_task_response.tasksInQueue:
                        typer.echo(f"{queued_task_item.public_id} - {queued_task_item.frontend_status.status_translation}")
                return run_task_response.task
            else:
                typer.echo(f'The task failed with an error: {run_task_response.message}')
                raise typer.Exit(1)
        else:
            typer.echo("The task failed with an error")
            raise typer.Exit(1)

    @error_handler()
    def cancel_task(self, task_public_id: str):
        cancel_result = self.__task_api_client.cancel_task(
            task_public_id=task_public_id,
        )

        if cancel_result.isSuccess:
            typer.echo(f'Task {task_public_id} has been canceled')
        else:
            typer.echo(f'Task {task_public_id} could not be canceled: {cancel_result.message}')

    def print_task_list(self, project_public_id: Optional[str], project_slug: Optional[str], row, page):
        if not project_slug and not project_public_id:
            project_config: ProjectConfig = self.__config_provider.read_project_config()
            if not project_config:
                typer.echo(
                    __("Provide the project unique ID or run this command from within an initialized project directory"))
                raise typer.Exit(1)
            project_public_id = project_config.public_id

        self.print(
            func_get_data=self.get_project_task_list,
            func_special_params={
                'project_public_id': project_public_id,
                'project_slug': project_slug,
            },
            mapper=TaskMapper(),
            headers=list(map(lambda x: x.alias, TaskEntity.model_fields.values())),
            row=row,
            page=page,
            max_col_width=[100, 100, 100, 100, 100, 100, 100, 100],
            show_index="never",
        )

    @error_handler()
    def get_project_task_list(
            self,
            project_public_id: Optional[str],
            project_slug: Optional[str],
            row: int = 5,
            page: int = 1,
    ) -> PaginatedEntityList[Task]:
        data: Optional[PaginatedEntityList[Task]] = self.__task_api_client.get_task_list_for_project(
            project_public_id=project_public_id,
            project_slug=project_slug,
            page=page,
            limit=row,
        )

        return data


from enum import Enum
from typing import List

# TODO reduce usage of statuses to zero
class TaskStatus(str, Enum):
    NEW: str = 'NEW'
    SCHEDULED: str = 'SCHEDULED'
    RUNNING: str = 'RUNNING'
    FINISHED: str = 'FINISHED'
    CANCELING: str = 'CANCELING'
    CANCELED: str = 'CANCELED'
    FAILED: str = 'FAILED'



from typing import Optional, Dict, Any, List

from pydantic import Field, BaseModel, ConfigDict

from thestage.services.clients.thestage_api.dtos.frontend_status import FrontendStatusDto
from thestage.services.clients.thestage_api.dtos.base_response import TheStageBasePaginatedResponse, \
    TheStageBaseResponse
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList
from thestage.instance.dto.enum.gpu_name import InstanceGpuType
from thestage.instance.dto.enum.cpu_type import InstanceCpuType


class InstanceRentedCreateRequest(BaseModel):
    slug: Optional[str] = Field(None, alias='slug')
    gpuName: Optional[str] = Field(None, alias='gpuName')
    cloudProviderName: Optional[str] = Field(None, alias='cloudProviderName')

    # optional prompted fields 'direct user input' values are str to avoid type casting errors
    gpuCount: Optional[str] = Field(None, alias='gpuCount')
    diskSizeGb: Optional[str] = Field(None, alias='diskSizeGb')

    forceStartRent: Optional[bool] = Field(None, alias='forceStartRent')
    concreteInstanceOfferPublicId: Optional[str] = Field(None, alias='concreteInstanceOfferPublicId')


class ConcreteInstanceOffer(BaseModel):
    publicId: Optional[str] = Field(None, alias='publicId', title='ID')
    # offerName: Optional[str] = Field(None, alias='offerName', title='Offer Name')
    providerAndRegion: Optional[str] = Field(None, alias='providerAndRegion', title='Provider and Region')
    gpuInfo: Optional[str] = Field(None, alias='gpuInfo', title='GPU')
    cpuInfo: Optional[str] = Field(None, alias='cpuInfo', title='CPU')
    diskSizeGb: Optional[int] = Field(None, alias='diskSizeGb', title='Disk size, GB')
    ramSizeGb: Optional[int] = Field(None, alias='ramSizeGb', title='RAM size, GB')
    pricePerHour: Optional[str] = Field(None, alias='pricePerHour', title='Price per hour')


class InstanceRentedCreateResponse(TheStageBaseResponse):
    instanceRentedPublicId: Optional[str] = Field(None, alias='instanceRentedPublicId')
    concreteInstanceOffersToSelect: Optional[List[ConcreteInstanceOffer]] = Field(None, alias='concreteInstanceOffersToSelect')
    selectedConcreteInstanceOffer: Optional[ConcreteInstanceOffer] = Field(None, alias='selectedConcreteInstanceOffer')

    missingFieldName: Optional[str] = Field(None, alias='missingFieldName')
    missingFieldDisplayName: Optional[str] = Field(None, alias='missingFieldDisplayName')
    missingFieldOptions: Optional[Dict[str, str]] = Field(None, alias='missingFieldOptions')



import traceback
from typing import Any, Callable

import requests
import typer
from click.exceptions import Exit, Abort
from git import GitCommandError
from paramiko.ssh_exception import PasswordRequiredException

from thestage.config.business import config_storage
from thestage.config.env_base import THESTAGE_API_URL
from thestage.exceptions.file_system_exception import FileSystemException
from thestage.exceptions.remote_server_exception import RemoteServerException
from thestage.i18n.translation import __
from thestage.exceptions.git_access_exception import GitAccessException
from thestage.exceptions.auth_exception import AuthException
from thestage.exceptions.business_logic_exception import BusinessLogicException
from thestage.exceptions.config_exception import ConfigException
from thestage.helpers.logger.app_logger import app_logger
from thestage.services.clients.thestage_api.core.http_client_exception import HttpClientException


def error_handler() -> Callable:
    def wrap(f):
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                result = f(*args, **kwargs)
                return result
            except AuthException as e1:
                typer.echo(__('Authentication failed: update access token'))
                app_logger.error(f'{traceback.format_exc()}')
                raise typer.Exit(1)
            except BusinessLogicException as e2:
                typer.echo(__('Business logic error encountered: contact TheStage AI team'))
                app_logger.error(f'{traceback.format_exc()}')
                raise typer.Exit(1)
            except ConfigException as e3:
                typer.echo(__(
                    'Configuration error encountered: %error_message%',
                    {
                        'error_message': e3.get_message()
                    }
                ))
                app_logger.error(f'{traceback.format_exc()}')
                raise typer.Exit(1)
            except FileSystemException as e4:
                typer.echo(__(
                    "File system error encountered: %error_message%",
                    {
                        'error_message': e4.get_message()
                    }
                ))
                app_logger.error(f'{traceback.format_exc()}')
                raise typer.Exit(1)
            except HttpClientException as e5:
                typer.echo(__(
                    f"TheStage server error: %error_message%",

                    {
                        'stage_url': THESTAGE_API_URL,
                        'error_message': e5.get_message()
                    }
                ))
                try:
                    app_logger.error(f"Connection error to {config_storage.APP_CONFIG.main.thestage_api_url} - {e5}")
                except Exception:
                    pass
                app_logger.error(f'{traceback.format_exc()}')
                raise typer.Exit(1)
            except GitAccessException as e6:
                typer.echo(e6.get_message())
                typer.echo(e6.get_dop_message())
                typer.echo(__(
                    "Visit %git_url% to accept the invitation or check your access to the repository",
                    {
                        'git_url': e6.get_url(),
                    }
                ))
                app_logger.error(f'{traceback.format_exc()}')
                raise typer.Exit(1)
            except GitCommandError as e7:
                typer.echo(f'Git command error encountered: {e7.stdout} \n {e7.stderr} (status: {e7.status})')
                app_logger.error(f'{traceback.format_exc()}')
                raise typer.Exit(1)
            except RemoteServerException as e8:
                typer.echo(__(
                    'Error connecting to server or Docker container at %ip_address% as %username%',
                    {
                        'ip_address': e8.ip_address,
                        'username': e8.username,
                    }))
                app_logger.error(f'Error connecting to server or Docker container at {e8.ip_address} as {e8.username} - {e8}')
                app_logger.error(f'{traceback.format_exc()}')
                raise typer.Exit(1)
            except Abort as e9:
                app_logger.error(f'{traceback.format_exc()}')
                raise typer.Exit(1)
            except requests.exceptions.ConnectionError as e10:
                # TODO we don't know for sure if it is Thestage connection error - throw appropriate exception (on token validation?)
                typer.echo("Connection error")
                app_logger.error(f'{traceback.format_exc()}')
                raise typer.Exit(1)
            except PasswordRequiredException as e11:
                # technically we can use encrypted keys but dealing with passwords is big bs
                typer.echo("Provided key requires password. Use non-encrypted key.")
                app_logger.error(f'{traceback.format_exc()}')
                raise typer.Exit(1)
            except Exception as e100:
                if isinstance(e100, Exit):
                    raise e100
                else:
                    typer.echo(__('Undefined error occurred'))
                    # TODO send all exceptions to backend?
                    app_logger.error(f'{traceback.format_exc()}')
                    print(f'{traceback.format_exc()}')
                    raise typer.Exit(1)
        return wrapper
    return wrap



from typing import Optional, Dict, Any, List

from pydantic import Field

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBasePaginatedResponse, \
    TheStageBaseResponse


class InstanceRentedTerminateResponse(TheStageBaseResponse):
    terminatedInstanceRentedPublicIds: Optional[List[str]] = Field(None, alias='terminatedInstanceRentedPublicIds')
    failedInstanceRentedPublicIds: Optional[List[str]] = Field(None, alias='failedInstanceRentedPublicIds')



from typing import Optional

from thestage.config.business.app_config_service import AppConfigService
from thestage.config.business.config_provider import ConfigProvider
from thestage.config.business.validation_service import ValidationService
from thestage.connect.business.connect_service import ConnectService
from thestage.connect.business.remote_server_service import RemoteServerService
from thestage.connect.communication.connect_api_client import ConnectApiClient
from thestage.docker_container.business.container_service import ContainerService
from thestage.docker_container.communication.docker_container_api_client import DockerContainerApiClient
from thestage.git.communication.git_client import GitLocalClient
from thestage.inference_model.business.inference_model_service import InferenceModelService
from thestage.inference_model.communication.inference_model_api_client import InferenceModelApiClient
from thestage.inference_simulator.business.inference_simulator_service import InferenceSimulatorService
from thestage.inference_simulator.communication.inference_simulator_api_client import InferenceSimulatorApiClient
from thestage.instance.business.instance_service import InstanceService
from thestage.instance.communication.instance_api_client import InstanceApiClient
from thestage.logging.business.logging_service import LoggingService
from thestage.logging.communication.logging_api_client import LoggingApiClient
from thestage.project.business.project_service import ProjectService
from thestage.project.communication.project_api_client import ProjectApiClient
from thestage.services.clients.thestage_api.core.api_client_core import TheStageApiClientCore
from thestage.services.filesystem_service import FileSystemService
from thestage.task.business.task_service import TaskService
from thestage.task.communication.task_api_client import TaskApiClient


class ServiceFactory:
    __git_local_client: Optional[GitLocalClient] = None
    __file_system_service: Optional[FileSystemService] = None
    __config_provider: Optional[ConfigProvider] = None
    __task_api_client: Optional[TaskApiClient] = None
    __project_api_client: Optional[ProjectApiClient] = None
    __inference_model_api_client: Optional[InferenceModelApiClient] = None
    __inference_simulator_api_client: Optional[InferenceSimulatorApiClient] = None
    __instance_api_client: Optional[InstanceApiClient] = None
    __docker_container_api_client: Optional[DockerContainerApiClient] = None
    __connect_api_client: Optional[ConnectApiClient] = None
    __logging_api_client: Optional[LoggingApiClient] = None

    def get_validation_service(self) -> ValidationService:
        config_provider = self.get_config_provider()
        core_client = TheStageApiClientCore(url=config_provider.get_config().main.thestage_api_url)
        return ValidationService(
            core_client=core_client,
            config_provider=config_provider,
        )

    def get_instance_service(self) -> InstanceService:
        return InstanceService(
            instance_api_client=self.get_instance_api_client(),
            remote_server_service=self.get_remote_server_service(),
            config_provider=self.get_config_provider(),
        )

    def get_container_service(self) -> ContainerService:
        return ContainerService(
            docker_container_api_client=self.get_docker_container_api_client(),
            remote_server_service=self.get_remote_server_service(),
            file_system_service=self.get_file_system_service(),
            config_provider=self.get_config_provider(),
            connect_api_client=self.get_connect_api_client(),
        )

    def get_connect_service(self) -> ConnectService:
        return ConnectService(
            instance_api_client=self.get_instance_api_client(),
            connect_api_client=self.get_connect_api_client(),
            instance_service=self.get_instance_service(),
            container_service=self.get_container_service(),
            logging_service=self.get_logging_service(),
        )

    def get_project_service(self) -> ProjectService:
        return ProjectService(
            task_api_client=self.get_task_api_client(),
            docker_container_api_client=self.get_docker_container_api_client(),
            project_api_client=self.get_project_api_client(),
            remote_server_service=self.get_remote_server_service(),
            file_system_service=self.get_file_system_service(),
            git_local_client=self.get_git_local_client(),
            config_provider=self.get_config_provider(),
        )

    def get_task_service(self) -> TaskService:
        return TaskService(
            docker_container_api_client=self.get_docker_container_api_client(),
            project_service=self.get_project_service(),
            task_api_client=self.get_task_api_client(),
            config_provider=self.get_config_provider(),
            git_local_client=self.get_git_local_client(),
            file_system_service=self.get_file_system_service(),
        )

    def get_inference_simulator_service(self) -> InferenceSimulatorService:
        return InferenceSimulatorService(
            inference_simulator_api_client=self.get_inference_simulator_api_client(),
            inference_model_api_client=self.get_inference_model_api_client(),
            config_provider=self.get_config_provider(),
            git_local_client=self.get_git_local_client(),
            project_service=self.get_project_service(),
        )

    def get_inference_model_service(self) -> InferenceModelService:
        return InferenceModelService(
            inference_model_api_client=self.get_inference_model_api_client(),
            config_provider=self.get_config_provider(),
            project_service=self.get_project_service(),
        )

    def get_remote_server_service(self) -> RemoteServerService:
        return RemoteServerService(
            file_system_service=self.get_file_system_service(),
            config_provider=self.get_config_provider(),
        )

    def get_app_config_service(self) -> AppConfigService:
        return AppConfigService(
            config_provider=self.get_config_provider(),
            validation_service=self.get_validation_service(),
        )

    def get_logging_service(self) -> LoggingService:
        return LoggingService(
            logging_api_client=self.get_logging_api_client(),
            docker_container_api_client=self.get_docker_container_api_client(),
            task_api_client=self.get_task_api_client(),
            inference_simulator_api_client=self.get_inference_simulator_api_client(),
        )

    def get_task_api_client(self) -> TaskApiClient:
        if not self.__task_api_client:
            self.__task_api_client = TaskApiClient(config_provider=self.get_config_provider())
        return self.__task_api_client

    def get_project_api_client(self) -> ProjectApiClient:
        if not self.__project_api_client:
            self.__project_api_client = ProjectApiClient(config_provider=self.get_config_provider())
        return self.__project_api_client

    def get_inference_model_api_client(self) -> InferenceModelApiClient:
        if not self.__inference_model_api_client:
            self.__inference_model_api_client = InferenceModelApiClient(config_provider=self.get_config_provider())
        return self.__inference_model_api_client

    def get_inference_simulator_api_client(self) -> InferenceSimulatorApiClient:
        if not self.__inference_simulator_api_client:
            self.__inference_simulator_api_client = InferenceSimulatorApiClient(config_provider=self.get_config_provider())
        return self.__inference_simulator_api_client

    def get_instance_api_client(self) -> InstanceApiClient:
        if not self.__instance_api_client:
            self.__instance_api_client = InstanceApiClient(config_provider=self.get_config_provider())
        return self.__instance_api_client

    def get_docker_container_api_client(self) -> DockerContainerApiClient:
        if not self.__docker_container_api_client:
            self.__docker_container_api_client = DockerContainerApiClient(config_provider=self.get_config_provider())
        return self.__docker_container_api_client

    def get_connect_api_client(self) -> ConnectApiClient:
        if not self.__connect_api_client:
            self.__connect_api_client = ConnectApiClient(config_provider=self.get_config_provider())
        return self.__connect_api_client

    def get_logging_api_client(self) -> LoggingApiClient:
        if not self.__logging_api_client:
            self.__logging_api_client = LoggingApiClient(config_provider=self.get_config_provider())
        return self.__logging_api_client

    def get_git_local_client(self):
        if not self.__git_local_client:
            self.__git_local_client = GitLocalClient(file_system_service=self.get_file_system_service())
        return self.__git_local_client

    def get_file_system_service(self) -> FileSystemService:
        if not self.__file_system_service:
            self.__file_system_service = FileSystemService()
        return self.__file_system_service

    def get_config_provider(self) -> ConfigProvider:
        if not self.__config_provider:
            self.__config_provider = ConfigProvider(self.get_file_system_service())
        return self.__config_provider


from typing import Optional

from pydantic import ConfigDict, BaseModel, Field

from thestage.global_dto.enums.order_direction_type import OrderDirectionType


class EntityFilterRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    orderByField: Optional[str] = Field(None, alias='orderByField')
    orderByDirection: Optional[OrderDirectionType] = Field(None, alias='orderByDirection')
    page: Optional[int] = Field(None, alias='page')
    limit: Optional[int] = Field(None, alias='limit')




from pydantic import Field, ConfigDict

from thestage.inference_model.dto.inference_model import InferenceModel
from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList


class InferenceSimulatorModelListForProjectResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)

    inferenceSimulatorModels: PaginatedEntityList[InferenceModel] = Field(None, alias='inferenceSimulatorModels')



from typing import Optional, Dict

from pydantic import BaseModel, Field, ConfigDict


class DockerContainerMappings(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    directory_mappings: Dict[str, str] = Field(default_factory=dict, alias='directoryMappings')
    port_mappings: Dict[str, str] = Field(default_factory=dict, alias='portMappings')


class DockerContainerCreateRequest(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    instance_rented_public_id: Optional[str] = Field(None, alias='instanceRentedPublicId')
    instance_rented_slug: Optional[str] = Field(None, alias='instanceRentedSlug')
    selfhosted_instance_public_id: Optional[str] = Field(None, alias='selfhostedInstancePublicId')
    selfhosted_instance_slug: Optional[str] = Field(None, alias='selfhostedInstanceSlug')
    project_public_id: Optional[str] = Field(..., alias='projectPublicId')
    project_slug: Optional[str] = Field(..., alias='projectSlug')
    container_slug: str = Field(..., alias='containerSlug')
    docker_image_id: int = Field(..., alias='dockerImageId')
    assigned_cpus: Optional[int] = Field(None, alias='assignedCpus')
    is_unlimited_cpus: bool = Field(False, alias='isUnlimitedCpus')
    assigned_gpus: Optional[int] = Field(None, alias='assignedGpus')
    is_unlimited_gpus: bool = Field(False, alias='isUnlimitedGpus')
    mappings: DockerContainerMappings = Field(default_factory=DockerContainerMappings, alias='mappings')



from enum import Enum

# TODO reduce usage of statuses to zero
class SelfhostedBusinessStatus(str, Enum):
    AWAITING_CONFIGURATION: str = 'AWAITING_CONFIGURATION'
    RUNNING: str = 'RUNNING'
    UNREACHABLE_DAEMON: str = 'UNREACHABLE_DAEMON'
    DELETED: str = 'DELETED'
    UNKNOWN: str = 'UNKNOWN'
    ALL: str = 'ALL'



from typing import Optional

from thestage.docker_container.dto.container_response import DockerImageDto
from thestage.docker_container.dto.image_entity import DockerImageEntity
from thestage.services.abstract_mapper import AbstractMapper


class ImageMapper(AbstractMapper):
    def build_entity(self, item: DockerImageDto) -> Optional[DockerImageEntity]:
        if not item:
            return None

        return DockerImageEntity(
            image=item.image or '',
            tag=item.tag or '',
        )



from typing import Optional

from thestage.instance.dto.rented_instance import RentedInstanceEntity
from thestage.instance.dto.instance_rented_response import InstanceRentedDto
from thestage.services.abstract_mapper import AbstractMapper


class RentedInstanceMapper(AbstractMapper):

    def build_entity(self, item: InstanceRentedDto) -> Optional[RentedInstanceEntity]:
        if not item:
            return None

        return RentedInstanceEntity(
            slug=item.slug if item.slug else '',
            public_id=item.public_id if item.public_id else '',
            cpu_type=item.cpu_type if item.cpu_type else '',
            gpu_type=item.gpu_type if item.gpu_type else '',
            cpu_cores=str(item.cpu_cores) if item.cpu_cores else '',
            ip_address=item.ip_address if item.ip_address else '',
            status=item.frontend_status.status_translation if item.frontend_status else '',
        )



from typing import Optional

from pydantic import Field, ConfigDict, BaseModel


class UserLogsQueryRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    containerPublicId: Optional[str] = Field(None, alias='containerPublicId')
    taskPublicId: Optional[str] = Field(None, alias='taskPublicId')
    inferenceSimulatorPublicId: Optional[str] = Field(None, alias='inferenceSimulatorPublicId')
    limit: Optional[int] = Field(None, alias='limit')
    ascendingOrder: Optional[bool] = Field(None, alias='ascendingOrder')



from typing import Optional

from pydantic import Field, ConfigDict, BaseModel


class ProjectConfig(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    slug: str = Field(None, alias='slug')
    public_id: str = Field(None, alias='public_id')
    git_repository_url: str = Field(None, alias='git_repository_url')
    deploy_key_path: str = Field(None, alias='deploy_key_path')
    default_container_public_id: Optional[str] = Field(None, alias='default_container_public_id')
    prompt_for_default_container: bool = Field(True, alias='prompt_for_default_container')



import traceback

from thestage.cli_command_helper import get_command_group_help_panel
from thestage.helpers.logger.app_logger import app_logger, get_log_path_from_os
from thestage.services.clients.thestage_api.core.api_client_core import TheStageApiClientCore
from thestage.services.service_factory import ServiceFactory
from rich import print


def main():
    service_factory = ServiceFactory()
    config_provider = service_factory.get_config_provider()
    config = config_provider.build_config()

    try:
        try:
            api_client = TheStageApiClientCore(url=config.main.thestage_api_url)
            token_info = api_client.validate_token(config.main.thestage_auth_token)
            config_provider.update_allowed_commands_and_is_token_valid(validate_token_response=token_info)
        except Exception as e:
            config_provider.update_allowed_commands_and_is_token_valid(validate_token_response=None)
            app_logger.error(f'{traceback.format_exc()}')
            print(f'Application logs path: {str(get_log_path_from_os())}')

        base_controller = build_app()
        base_controller.app()
    except KeyboardInterrupt:
        print('THESTAGE: Keyboard Interrupt')


def build_app():
    from thestage.controllers import base_controller
    from thestage.config.communication import config_command
    from thestage.project.communication import project_command
    from thestage.inference_simulator.communication import inference_simulator_command
    from thestage.inference_model.communication import inference_model_command
    from thestage.task.communication import task_command
    from thestage.instance.communication import instance_command
    from thestage.docker_container.communication import docker_command
    from thestage.docker_container.communication import image_command

    project_command.app.add_typer(
        inference_simulator_command.app,
        name="inference-simulator",
        rich_help_panel=get_command_group_help_panel()
    )
    project_command.app.add_typer(
        inference_model_command.app,
        name="model",
        rich_help_panel=get_command_group_help_panel()
    )
    project_command.app.add_typer(
        task_command.app,
        name="task",
        rich_help_panel=get_command_group_help_panel()
    )

    docker_command.app.add_typer(
        image_command.app,
        name="image",
        rich_help_panel=get_command_group_help_panel()
    )

    base_controller.app.add_typer(
        project_command.app,
        name="project",
        rich_help_panel=get_command_group_help_panel()
    )

    base_controller.app.add_typer(
        docker_command.app,
        name="container",
        rich_help_panel=get_command_group_help_panel()
    )
    base_controller.app.add_typer(
        instance_command.app,
        name="instance",
        rich_help_panel=get_command_group_help_panel()
    )
    base_controller.app.add_typer(
        config_command.app,
        name="config",
        rich_help_panel=get_command_group_help_panel()
    )

    return base_controller


from thestage.services.clients.thestage_api.dtos.base_response import BaseResponseWarning
from rich import print


def print_warning(warning: BaseResponseWarning):
    colored_warning = f"[{warning.colorHex}]{warning.message}[/]"
    print(colored_warning)



from typing import Optional, List

import typer
from paramiko.dsskey import DSSKey
from paramiko.ecdsakey import ECDSAKey
from paramiko.ed25519key import Ed25519Key
from paramiko.pkey import PKey
from paramiko.rsakey import RSAKey
from paramiko.ssh_exception import SSHException, PasswordRequiredException


def parse_private_key(filename: str) -> Optional[PKey]:
    error_output: List[str] = []
    pkey: Optional[PKey] = None
    pkey_class: Optional[str] = None

    for pkey_class_item in (RSAKey, ECDSAKey, DSSKey, Ed25519Key):
        try:
            key = pkey_class_item.from_private_key_file(filename=filename)
            pkey = key
            pkey_class = pkey_class_item.__name__
            break
        except PasswordRequiredException as e1:
            raise e1
        except SSHException as e2:
            error_output.append(f'[{pkey_class_item.__name__}]: {str(e2)}')
        except Exception as e3:
            error_output.append(f'[{pkey_class_item.__name__}]: Unexpected error: {e3.__class__.__name__}: {str(e3)}')

    if pkey is None:
        for line in error_output:
            typer.echo(line)

    if pkey_class and pkey_class == 'DSSKey':
        typer.echo('DSA keys are not supported')
        return None

    return pkey



from typing import List, Optional

from pydantic import Field, ConfigDict

from thestage.docker_container.dto.container_response import DockerImageDto
from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList


class DockerImageListResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)

    paginatedList: PaginatedEntityList[DockerImageDto] = Field(None, alias='paginatedList')



from typing import Optional, List, Dict

from thestage.config.business.config_provider import ConfigProvider
from thestage.global_dto.enums.order_direction_type import OrderDirectionType
from thestage.helpers.error_handler import error_handler
from thestage.services.clients.thestage_api.core.api_client_core import TheStageApiClientCore
from thestage.services.clients.thestage_api.dtos.entity_filter_request import EntityFilterRequest
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList

from thestage.inference_simulator.dto.inference_simulator import InferenceSimulator
from thestage.inference_simulator.dto.inference_simulator_list_request import InferenceSimulatorListRequest
from thestage.inference_simulator.dto.inference_simulator_list_response import InferenceSimulatorListResponse
from thestage.inference_simulator.dto.start_inference_simulator_request import StartInferenceSimulatorRequest
from thestage.inference_simulator.dto.start_inference_simulator_response import StartInferenceSimulatorResponse
from thestage.inference_simulator.dto.inference_simulator_response import InferenceSimulatorStatusMapperResponse
from thestage.inference_simulator.dto.get_inference_simulator_request import GetInferenceSimulatorRequest
from thestage.inference_simulator.dto.get_inference_simulator_response import GetInferenceSimulatorResponse


class InferenceSimulatorApiClient(TheStageApiClientCore):
    def __init__(self, config_provider: ConfigProvider):
        super().__init__(url=config_provider.get_config().main.thestage_api_url)
        self.__config_provider = config_provider

    def get_inference_simulator_list(
            self,
            project_public_id: Optional[str],
            project_slug: Optional[str],
            statuses: List[str] = [],
            page: int = 1,
            limit: int = 10,
    ) -> Optional[PaginatedEntityList[InferenceSimulator]]:
        request = InferenceSimulatorListRequest(
            projectPublicId=project_public_id,
            projectSlug=project_slug,
            statuses=statuses,
            entityFilterRequest=EntityFilterRequest(
                orderByField="createdAt",
                orderByDirection=OrderDirectionType.DESC,
                page=page,
                limit=limit,
            ),
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/inference-simulator/list',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )
        result = InferenceSimulatorListResponse.model_validate(response) if response else None
        return result.inferenceSimulators if result and result.isSuccess else None

    def start_project_inference_simulator(
            self,
            project_public_id: str,
            commit_hash: Optional[str] = None,
            rented_instance_public_id: Optional[str] = None,
            rented_instance_slug: Optional[str] = None,
            self_hosted_instance_public_id: Optional[str] = None,
            self_hosted_instance_slug: Optional[str] = None,
            inference_dir: Optional[str] = None,
            is_skip_installation: Optional[bool] = False,
    ) -> Optional[StartInferenceSimulatorResponse]:
        request = StartInferenceSimulatorRequest(
            projectPublicId=project_public_id,
            commitHash=commit_hash,
            instanceRentedPublicId=rented_instance_public_id,
            instanceRentedSlug=rented_instance_slug,
            selfhostedInstancePublicId=self_hosted_instance_public_id,
            selfhostedInstanceSlug=self_hosted_instance_slug,
            inferenceDir=inference_dir,
            isSkipInstallation=is_skip_installation,
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/project/inference-simulator/create',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        return StartInferenceSimulatorResponse.model_validate(response) if response else None

    def get_inference_simulator_business_status_map(self) -> Optional[Dict[str, str]]:
        response = self._request(
            method='POST',
            url='/user-api/v1/inference-simulator/status-localized-mapping',
            data=None,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        data = InferenceSimulatorStatusMapperResponse.model_validate(response) if response else None

        return data.inference_simulator_status_map if data else None

    @error_handler()
    def get_inference_simulator(
            self,
            public_id: Optional[str] = None,
            slug: Optional[str] = None,
    ) -> Optional[GetInferenceSimulatorResponse]:
        request = GetInferenceSimulatorRequest(
            publicId=public_id,
            slug=slug,
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/inference-simulator/get',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )
        return GetInferenceSimulatorResponse.model_validate(response) if response else None


from typing import Optional

from pydantic import Field, ConfigDict, BaseModel


class PushInferenceSimulatorModelRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    inferenceSimulatorPublicId: Optional[str] = Field(None, alias='inferenceSimulatorPublicId')
    inferenceSimulatorSlug: Optional[str] = Field(None, alias='inferenceSimulatorSlug')



from typing import Optional, List, Dict

from thestage.config.business.config_provider import ConfigProvider
from thestage.global_dto.enums.order_direction_type import OrderDirectionType
from thestage.project.dto.delete_project_request import DeleteProjectRequest
from thestage.project.dto.delete_project_response import DeleteProjectResponse
from thestage.project.dto.list_project_request import ProjectListRequest
from thestage.project.dto.list_project_response import ProjectListResponse, ProjectBusinessStatusMapperResponse
from thestage.project.dto.modify_project_instance_assignment_request import ModifyProjectInstanceAssignmentRequest
from thestage.project.dto.modify_project_instance_assignment_response import ModifyProjectInstanceAssignmentResponse
from thestage.services.clients.thestage_api.core.api_client_core import TheStageApiClientCore

from thestage.project.dto.project_response import ProjectDto, ProjectViewResponse
from thestage.project.dto.get_deploy_ssh_key_request import ProjectGetDeploySshKeyRequest
from thestage.project.dto.get_deploy_ssh_key_response import ProjectGetDeploySshKeyResponse
from thestage.services.clients.thestage_api.dtos.entity_filter_request import EntityFilterRequest
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList


class ProjectApiClient(TheStageApiClientCore):
    def __init__(self, config_provider: ConfigProvider):
        super().__init__(url=config_provider.get_config().main.thestage_api_url)
        self.__config_provider = config_provider

    def get_project(self, slug: Optional[str], public_id: Optional[str], default_container_public_id: Optional[str]) -> Optional[ProjectViewResponse]:
        data = {
            "projectSlug": slug,
            "projectPublicId": public_id,
            "defaultContainerPublicId": default_container_public_id
        }

        response = self._request(
            method='POST',
            url='/user-api/v2/project/view',
            data=data,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = ProjectViewResponse.model_validate(response) if response else None
        return result if result and result.isSuccess else None

    def get_project_deploy_ssh_key(self, public_id: str) -> str:
        request = ProjectGetDeploySshKeyRequest(
            projectPublicId=public_id,
        )

        response = self._request(
            method='POST',
            url='/user-api/v1/project/get-deploy-ssh-key',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = ProjectGetDeploySshKeyResponse.model_validate(response) if response else None
        return result.privateKey if result and result.isSuccess else None

    def get_project_list(
            self,
            page: int = 1,
            limit: int = 10,
            statuses: List[str] = None,
            rented_instance_public_id: Optional[str] = None,
            rented_instance_slug: Optional[str] = None,
            self_hosted_instance_public_id: Optional[str] = None,
            self_hosted_instance_slug: Optional[str] = None,
    ) -> PaginatedEntityList[ProjectDto]:
        request = ProjectListRequest(
            statuses=statuses,
            rentedInstancePublicId=rented_instance_public_id,
            rentedInstanceSlug=rented_instance_slug,
            selfHostedInstancePublicId=self_hosted_instance_public_id,
            selfHostedInstanceSlug=self_hosted_instance_slug,
            entityFilterRequest=EntityFilterRequest(
                orderByField="createdAt",
                orderByDirection=OrderDirectionType.DESC,
                page=page,
                limit=limit,
            ),
        )

        response = self._request(
            method='POST',
            url='/user-api/v1/project/list',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = ProjectListResponse.model_validate(response) if response else None
        return result.paginatedList if result and result.isSuccess else None


    def get_project_business_status_map(self) -> Optional[Dict[str, str]]:
        response = self._request(
            method='POST',
            url='/user-api/v1/project/status-localized-map',
            data=None,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        data = ProjectBusinessStatusMapperResponse.model_validate(response) if response else None
        return data.project_business_status_map if data else None


    def modify_project_instance_assignment(
            self,
            project_public_id,
            project_slug,
            rented_instance_public_id_to_assign,
            rented_instance_slug_to_assign,
            rented_instance_public_id_to_delete,
            rented_instance_slug_to_delete,
            self_hosted_instance_public_id_to_assign,
            self_hosted_instance_slug_to_assign,
            self_hosted_instance_public_id_to_delete,
            self_hosted_instance_slug_to_delete) -> Optional[ModifyProjectInstanceAssignmentResponse]:

        request = ModifyProjectInstanceAssignmentRequest(
            projectPublicId=project_public_id,
            projectSlug=project_slug,
            instanceRentedPublicIdsToAssign=rented_instance_public_id_to_assign,
            instanceRentedSlugsToAssign=rented_instance_slug_to_assign,
            instanceRentedPublicIdsToDelete=rented_instance_public_id_to_delete,
            instanceRentedSlugsToDelete=rented_instance_slug_to_delete,
            selfhostedInstancePublicIdsToAssign=self_hosted_instance_public_id_to_assign,
            selfhostedInstanceSlugsToAssign=self_hosted_instance_slug_to_assign,
            selfhostedInstancePublicIdsToDelete=self_hosted_instance_public_id_to_delete,
            selfhostedInstanceSlugsToDelete=self_hosted_instance_slug_to_delete,
        )

        response = self._request(
            method='POST',
            url='/user-api/v1/project/modify-instance-assignment',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        return ModifyProjectInstanceAssignmentResponse.model_validate(response) if response else None


    def delete_project(self, project_public_id, project_slug) -> Optional[DeleteProjectResponse]:
        request = DeleteProjectRequest(
            projectPublicId=project_public_id,
            projectSlug=project_slug,
        )

        response = self._request(
            method='POST',
            url='/user-api/v1/project/delete',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        return DeleteProjectResponse.model_validate(response) if response else None



from typing import Optional

from pydantic import Field, ConfigDict, BaseModel


class AddSshPublicKeyToInstanceRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    instanceRentedPublicId: str = Field(None, alias='instanceRentedPublicId')
    sshPublicKeyPublicId: str = Field(None, alias='sshPublicKeyPublicId')



from typing import Optional

from pydantic import Field, ConfigDict

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse


class DockerContainerCreateResult(TheStageBaseResponse):
    model_config = ConfigDict(populate_by_name=True)

    docker_container_public_id: Optional[str] = Field(None, alias='dockerContainerPublicId')



from typing import Optional

import httpx

from thestage.config.business.config_provider import ConfigProvider
from thestage.services.clients.thestage_api.core.api_client_core import TheStageApiClientCore
from thestage.logging.dto.log_polling_request import LogPollingRequest
from thestage.logging.dto.log_polling_response import LogPollingResponse
from thestage.logging.dto.user_logs_query_request import UserLogsQueryRequest
from thestage.logging.dto.user_logs_query_response import UserLogsQueryResponse


class LoggingApiClient(TheStageApiClientCore):
    def __init__(self, config_provider: ConfigProvider):
        super().__init__(url=config_provider.get_config().main.thestage_api_url)
        self.__config_provider = config_provider

    async def poll_logs_httpx(self, docker_container_public_id: Optional[str], last_log_timestamp: str,
                              last_log_id: str, task_public_id: Optional[str] = None,
                              inference_simulator_public_id: Optional[str] = None) -> Optional[LogPollingResponse]:
        request_headers = {'Content-Type': 'application/json'}
        token = self.__config_provider.get_config().main.thestage_auth_token
        if token: request_headers['Authorization'] = f"Bearer {token}"

        request = LogPollingRequest(
            inferenceSimulatorPublicId=inference_simulator_public_id,
            taskPublicId=task_public_id,
            dockerContainerPublicId=docker_container_public_id,
            lastLogTimestamp=last_log_timestamp,
            lastLogId=last_log_id
        )

        async with httpx.AsyncClient() as client:
            response = await client.post(
                url=f"{self._get_host()}/user-api/v2/logging/poll",
                headers=request_headers,
                json=request.model_dump(),
                timeout=3.5
            )

            return LogPollingResponse.model_validate(response.json()) if response else None


    def query_user_logs(self, limit: int, task_public_id: Optional[str] = None,
                        inference_simulator_public_id: Optional[str] = None,
                        container_public_id: Optional[str] = None) -> UserLogsQueryResponse:
        request = UserLogsQueryRequest(
            inferenceSimulatorPublicId=inference_simulator_public_id,
            taskPublicId=task_public_id,
            containerPublicId=container_public_id,
            limit=limit,
            ascendingOrder=False,
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/logging/query-user-logs',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = UserLogsQueryResponse.model_validate(response) if response else None
        return result


from typing import Optional, List

from pydantic import Field, ConfigDict

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.task.dto.task import Task


class RunTaskResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)

    task: Task = Field(None, alias='task')
    tasksInQueue: Optional[List[Task]] = Field(None, alias='tasksInQueue')



from typing import Optional

from pydantic import Field, BaseModel, ConfigDict

from thestage.docker_container.dto.enum.container_pending_action import DockerContainerAction

class DockerContainerActionRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    dockerContainerPublicId: Optional[str] = Field(None, alias='dockerContainerPublicId')
    dockerContainerSlug: Optional[str] = Field(None, alias='dockerContainerSlug')
    action: DockerContainerAction = Field(None, alias='action')



from typing import Optional, List, Dict

from thestage.config.business.config_provider import ConfigProvider
from thestage.instance.dto.instance_rented_create_api_dtos import InstanceRentedCreateRequest, InstanceRentedCreateResponse
from thestage.instance.dto.instance_rented_terminate_api_dtos import InstanceRentedTerminateResponse
from thestage.services.clients.thestage_api.core.api_client_core import TheStageApiClientCore
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList

from thestage.instance.dto.instance_rented_response import (
    InstanceRentedListResponse,
    InstanceRentedDto,
    InstanceRentedItemResponse,
    InstanceRentedBusinessStatusMapperResponse
)
from thestage.instance.dto.selfhosted_instance_response import (
    SelfHostedInstanceListResponse,
    SelfHostedInstanceDto,
    SelfHostedRentedItemResponse,
    SelfHostedRentedRentedBusinessStatusMapperResponse
)


class InstanceApiClient(TheStageApiClientCore):
    def __init__(self, config_provider: ConfigProvider):
        super().__init__(url=config_provider.get_config().main.thestage_api_url)
        self.__config_provider = config_provider

    def get_rented_instance_list(
            self,
            statuses: List[str],
            page: int = 1,
            limit: int = 10,
    ) -> PaginatedEntityList[InstanceRentedDto]:
        data = {
            #"statuses": [item.value for item in statuses],
            "entityFilterRequest": {
                "orderByField": "createdAt",
                "orderByDirection": "DESC",
                "page": page,
                "limit": limit
            },
        }

        if statuses:
            data['businessStatuses'] = statuses

        response = self._request(
            method='POST',
            url='/user-api/v3/instance-rented/list',
            data=data,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = InstanceRentedListResponse.model_validate(response) if response else None
        return result.paginated_list if result and result.paginated_list else ([], None)

    def get_rented_business_status_map(self) -> Optional[Dict[str, str]]:
        response = self._request(
            method='POST',
            url='/user-api/v2/instance-rented/business-status-localized-map',
            data=None,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        data = InstanceRentedBusinessStatusMapperResponse.model_validate(response) if response else None

        return data.instance_rented_business_status_map if data else None



    def get_rented_instance(
            self,
            instance_public_id: Optional[str] = None,
            instance_slug: Optional[str] = None,
    ) -> Optional[InstanceRentedDto]:
        if not instance_slug and not instance_public_id:
            return None

        data = {
            "instanceRentedPublicId": instance_public_id,
            "instanceRentedSlug": instance_slug,
        }

        response = self._request(
            method='POST',
            url='/user-api/v3/instance-rented/view',
            data=data,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        return InstanceRentedItemResponse.model_validate(response).instance_rented if response else None

    def get_selfhosted_instance(
            self,
            instance_public_id: Optional[str] = None,
            instance_slug: Optional[str] = None,
    ) -> Optional[SelfHostedInstanceDto]:
        if not instance_slug and not instance_public_id:
            return None

        data = {
            "selfhostedInstancePublicId": instance_public_id,
            "selfhostedInstanceSlug": instance_slug,
        }

        response = self._request(
            method='POST',
            url='/user-api/v3/self-hosted-instance/view',
            data=data,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        return SelfHostedRentedItemResponse.model_validate(response).selfhosted_instance if response else None

    def get_selfhosted_instance_list(
            self,
            statuses: List[str],
            page: int = 1,
            limit: int = 10,
    ) -> PaginatedEntityList[SelfHostedInstanceDto]:
        data = {
            "entityFilterRequest": {
                "orderByField": "createdAt",
                "orderByDirection": "DESC",
                "page": page,
                "limit": limit
            }
        }

        if statuses:
            data['businessStatuses'] = statuses

        response = self._request(
            method='POST',
            url='/user-api/v3/self-hosted-instance/list',
            data=data,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = SelfHostedInstanceListResponse.model_validate(response) if response else None
        return result.paginated_list if result and result.paginated_list else ([], None)

    def get_selfhosted_business_status_map(self) -> Optional[Dict[str, str]]:
        response = self._request(
            method='POST',
            url='/user-api/v2/self-hosted-instance/business-status-localized-map',
            data=None,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        data = SelfHostedRentedRentedBusinessStatusMapperResponse.model_validate(response) if response else None
        return data.selfhosted_instance_business_status_map if data else None


    def create_instance_rented(
            self,
            request: InstanceRentedCreateRequest,
    ) -> Optional[InstanceRentedCreateResponse]:

        response = self._request(
            method='POST',
            url='/user-api/v1/instance-rented/create',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        return InstanceRentedCreateResponse.model_validate(response) if response else None


    def terminate_instance_rented(
            self,
            instance_rented_public_id: Optional[str],
            instance_rented_slug: Optional[str]
    ) -> InstanceRentedTerminateResponse:

        data = {
            "instanceRentedPublicId": instance_rented_public_id,
            "instanceRentedSlug": instance_rented_slug,
        }

        response = self._request(
            method='POST',
            url='/user-api/v1/instance-rented/terminate',
            data=data,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = InstanceRentedTerminateResponse.model_validate(response) if response else None
        return result


    def restart_instance_rented(
            self,
            instance_rented_public_id: Optional[str],
            instance_rented_slug: Optional[str]
    ) -> InstanceRentedTerminateResponse:

        data = {
            "instanceRentedPublicId": instance_rented_public_id,
            "instanceRentedSlug": instance_rented_slug,
            "requestedPowerStatus": "REBOOT"
        }

        response = self._request(
            method='POST',
            url='/user-api/v1/instance-rented/switch-power-status',
            data=data,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = InstanceRentedTerminateResponse.model_validate(response) if response else None
        return result



from typing import Optional, List

from pydantic import Field, ConfigDict

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.logging.dto.log_message import LogMessage


class LogPollingResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)

    logs: Optional[List[LogMessage]] = Field(None, alias='logs')
    lastLogId: Optional[str] = Field(None, alias='lastLogId')
    lastLogTimestamp: Optional[str] = Field(None, alias='lastLogTimestamp')



from pathlib import Path

import click

from thestage.cli_command import CliCommand
from thestage.cli_command_helper import get_command_metadata, check_command_permission
from thestage.config.dto.config_entity import ConfigEntity
from thestage.global_dto.enums.yes_no_response import YesOrNoResponse

from thestage.i18n.translation import __
from thestage.helpers.logger.app_logger import app_logger, get_log_path_from_os
from thestage.connect.business.connect_service import ConnectService
from thestage.services.service_factory import ServiceFactory
from thestage.controllers.utils_controller import get_current_directory

import typer

app = typer.Typer(no_args_is_help=True, help=__("Manage configuration settings"))


@app.command(name='get', no_args_is_help=False, help=__("Display all configuration settings"), **get_command_metadata(CliCommand.CONFIG_GET))
def config_get():
    command_name = CliCommand.CONFIG_GET
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    config_provider = ServiceFactory().get_config_provider()
    config: ConfigEntity = config_provider.get_config()

    if not config:
        typer.echo(__('No configuration found'))
        raise typer.Exit(1)

    config_provider.save_config()

    typer.echo(__('THESTAGE TOKEN: %token%', {'token': config.main.thestage_auth_token or ''}))
    typer.echo(__('THESTAGE API LINK: %link%', {'link': config.main.thestage_api_url or ''}))

    if config.runtime.config_global_path:
        typer.echo(__('CONFIG PATH: %path%', {'path': str(config.runtime.config_global_path or '') + f'/config.json'}))

    typer.echo(__('APPLICATION LOGS PATH: %path%', {'path': str(get_log_path_from_os())}))

    raise typer.Exit(0)


@app.command(name='set', no_args_is_help=True, help=__("Update configuration settings"), **get_command_metadata(CliCommand.CONFIG_SET))
def config_set(
    token: str = typer.Option(
            None,
            "--access-token",
            "-t",
            help=__("Set or update access token"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONFIG_SET
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    service_factory = ServiceFactory()
    app_service = service_factory.get_app_config_service()

    if token:
        app_service.app_change_token(token=token)

    typer.echo('Configuration updated successfully')
    raise typer.Exit(0)


@app.command(name='clear', no_args_is_help=False, help=__("Clear configuration"), **get_command_metadata(CliCommand.CONFIG_CLEAR))
def config_clear():
    command_name = CliCommand.CONFIG_CLEAR
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    config_provider = ServiceFactory().get_config_provider()
    config_dir = config_provider.get_config().runtime.config_global_path
    config_provider.clear_config()
    typer.echo(f'Removed {config_dir}')

    raise typer.Exit(0)


@app.command(name='upload-ssh-key', no_args_is_help=True, help=__("Send your public SSH key to the platform and / or rented server instance"), **get_command_metadata(CliCommand.CONFIG_UPLOAD_SSH_KEY))
def upload_ssh_key(
        ssh_public_key: str = typer.Argument(
            help=__("Path to your public SSH key file or your public SSH key contents. OpenSSH key format is required."),
        ),
        instance_rented_public_id: str = typer.Option(
            None,
            "--instance-rented-id",
            "-rid",
            help=__("ID of your rented instance to add the key to (optional)"),
            is_eager=False,
        ),
        instance_rented_slug: str = typer.Option(
            None,
            "--instance-rented-name",
            "-rn",
            help=__("Name of your rented instance to add the key to (optional)"),
            is_eager=False,
        )
):
    command_name = CliCommand.CONFIG_UPLOAD_SSH_KEY
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [instance_rented_public_id, instance_rented_slug]) > 1:
        typer.echo("Provide a single identifier for rented instance - ID or name.")
        raise typer.Exit(1)

    service_factory = ServiceFactory()
    connect_service: ConnectService = service_factory.get_connect_service()

    is_path_provided_confirmed = False
    ssh_key_contents = ssh_public_key

    if ssh_public_key.startswith("/") or ssh_public_key.startswith("~") or ssh_public_key.endswith(".pub") or len(ssh_public_key) < 30:
        is_path_provided_confirmed = True

    ssh_key_path = Path(ssh_public_key).absolute()
    if is_path_provided_confirmed and not ssh_key_path.exists():
        typer.echo(f"No key was found at {ssh_key_path}")
        raise typer.Exit(1)

    if is_path_provided_confirmed or ssh_key_path.exists():
        if '.' not in ssh_key_path.name:
            proceed_with_no_extension: YesOrNoResponse = typer.prompt(
                text=f"File '{ssh_key_path.name}' probably contains a private key. Proceed?",
                show_choices=True,
                default=YesOrNoResponse.YES.value,
                type=click.Choice([r.value for r in YesOrNoResponse]),
                show_default=True,
            )
            if proceed_with_no_extension == YesOrNoResponse.NO:
                raise typer.Exit(0)
        ssh_key_contents = ssh_key_path.open("r").read()
        if 'private key-----' in ssh_key_contents.lower():
            typer.echo(f"{ssh_key_path} is identified as a private key. Provide a public SSH key.")
            raise typer.Exit(1)

    connect_service.upload_ssh_key(
        public_key_contents=ssh_key_contents,
        instance_public_id=instance_rented_public_id,
        instance_slug=instance_rented_slug,
    )

    raise typer.Exit(0)



from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class SelfHostedInstanceEntity(BaseModel):
    model_config = ConfigDict(
        use_enum_values=True,
        populate_by_name=True,
    )

    public_id: Optional[str] = Field(None, alias='ID')
    slug: Optional[str] = Field(None, alias='NAME')
    status: Optional[str] = Field(None, alias='STATUS')
    cpu_type: Optional[str] = Field(str, alias='CPU_TYPE')
    cpu_cores: Optional[int] = Field(None, alias='CPU_CORES')
    gpu_type: Optional[str] = Field(str, alias='GPU_TYPE')
    ip_address: Optional[str] = Field(None, alias='IP_ADDRESS')



from pathlib import Path
from typing import Optional, List

import typer

from thestage.cli_command import CliCommand
from thestage.cli_command_helper import get_command_metadata, check_command_permission
from thestage.controllers.utils_controller import validate_config_and_get_service_factory, get_current_directory
from thestage.helpers.logger.app_logger import app_logger
from thestage.i18n.translation import __
from thestage.inference_simulator.dto.get_inference_simulator_response import GetInferenceSimulatorResponse
from thestage.logging.business.logging_service import LoggingService

app = typer.Typer(no_args_is_help=True, help="Manage project inference simulators")

@app.command(name='run', no_args_is_help=True, help="Run inference simulator within the project", **get_command_metadata(CliCommand.PROJECT_INFERENCE_SIMULATOR_RUN))
def run_inference_simulator(
        rented_instance_public_id: Optional[str] = typer.Option(
            None,
            '--rented-instance-id',
            '-rid',
            help=__("Rented instance ID on which the inference simulator will run"),
            is_eager=False,
        ),
        rented_instance_slug: Optional[str] = typer.Option(
            None,
            '--rented-instance-name',
            '-rn',
            help=__("Rented instance name on which the inference simulator will run"),
            is_eager=False,
        ),
        self_hosted_instance_public_id: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-id',
            '-sid',
            help=__("Self-hosted instance ID on which the inference simulator will run"),
            is_eager=False,
        ),
        self_hosted_instance_slug: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-name',
            '-sn',
            help=__("Self-hosted instance name on which the inference simulator will run"),
            is_eager=False,
        ),
        commit_hash: Optional[str] = typer.Option(
            None,
            '--commit-hash',
            '-hash',
            help=__("Commit hash to use. By default, the current HEAD commit is used."),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory. By default, the current directory is used"),
            show_default=False,
            is_eager=False,
        ),
        enable_log_stream: Optional[bool] = typer.Option(
            True,
            "--no-logs",
            "-nl",
            help=__("Disable real-time log streaming"),
            is_eager=False,
        ),
        is_skip_installation: Optional[bool] = typer.Option(
            False,
            "--skip-installation",
            "-si",
            help=__("Skip installing dependencies from requirements.txt and install.sh"),
            is_eager=False,
        ),
        files_to_add: Optional[str] = typer.Option(
            None,
            "--files-add",
            "-fa",
            help=__("Files to add to the commit. You can add files by their relative path from the working directory with a comma as a separator."),
            is_eager=False,
        ),
        is_skip_auto_commit: Optional[bool] = typer.Option(
            False,
            "--skip-autocommit",
            "-sa",
            help=__("Skip automatic commit of the changes"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_INFERENCE_SIMULATOR_RUN
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    config = service_factory.get_config_provider().get_config()

    working_dir_path = Path(working_directory) if working_directory else Path(config.runtime.working_directory)
    inference_files = list(working_dir_path.rglob("inference.py"))
    if not inference_files:
        typer.echo("No inference.py file found in the project directory.")
        raise typer.Exit(1)
    elif len(inference_files) == 1:
        selected_inference = inference_files[0]
    else:
        choices = [str(path.relative_to(working_dir_path)) for path in inference_files]
        typer.echo("Multiple inference.py files found:")
        for idx, choice in enumerate(choices, start=1):
            typer.echo(f"{idx}) {choice}")
        choice_str = typer.prompt("Choose which inference.py to use")
        try:
            choice_index = int(choice_str)
        except ValueError:
            raise typer.BadParameter("Invalid input. Enter a number.")
        if not (1 <= choice_index <= len(choices)):
            raise typer.BadParameter("Choice out of range.")
        selected_inference = inference_files[choice_index - 1]

    relative_inference = selected_inference.relative_to(working_dir_path)
    parent_dir = relative_inference.parent
    if parent_dir == Path("../../controllers"):
        inference_dir = "/"
    else:
        inference_dir = f"{parent_dir.as_posix()}/"
    typer.echo(f"Selected inference file relative path: {inference_dir}")

    inference_simulator_service = service_factory.get_inference_simulator_service()

    inference_simulator = inference_simulator_service.project_run_inference_simulator(
        commit_hash=commit_hash,
        rented_instance_public_id=rented_instance_public_id,
        rented_instance_slug=rented_instance_slug,
        self_hosted_instance_public_id=self_hosted_instance_public_id,
        self_hosted_instance_slug=self_hosted_instance_slug,
        inference_dir=inference_dir,
        is_skip_installation=is_skip_installation,
        files_to_add=files_to_add,
        is_skip_auto_commit=is_skip_auto_commit,
    )

    if enable_log_stream:
        logging_service: LoggingService = service_factory.get_logging_service()

        logging_service.stream_inference_simulator_logs_with_controls(
            slug=inference_simulator.slug
        )
    raise typer.Exit(0)


@app.command(name='save-metadata', no_args_is_help=True, help="Save inference simulator metadata", **get_command_metadata(CliCommand.PROJECT_INFERENCE_SIMULATOR_SAVE_METADATA))
def get_and_save_inference_simulator_metadata(
        inference_simulator_public_id: Optional[str] = typer.Option(
            None,
            '--inference-simulator-id',
            '-isid',
            help=__("Inference simulator ID"),
            is_eager=False,
        ),
        inference_simulator_slug: Optional[str] = typer.Option(
            None,
            '--inference-simulator-name',
            '-isn',
            help=__("Inference simulator name"),
            is_eager=False,
        ),
        file_path: Optional[str] = typer.Option(
            None,
            "--file-path",
            "-fp",
            help=__("Full path to a new file. By default metadata is saved to the current directory as metadata.json"),
            show_default=False,
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_INFERENCE_SIMULATOR_SAVE_METADATA
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [inference_simulator_public_id, inference_simulator_slug]) != 1:
        typer.echo("Provide a single identifier for inference simulator - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    inference_simulator_service = service_factory.get_inference_simulator_service()

    inference_simulator_service.project_get_and_save_inference_simulator_metadata(
        file_path=file_path,
        inference_simulator_public_id=inference_simulator_public_id,
        inference_simulator_slug=inference_simulator_slug,
    )

    raise typer.Exit(0)


@app.command(name='push', no_args_is_help=True, help="Push inference simulator to model registry", **get_command_metadata(CliCommand.PROJECT_INFERENCE_SIMULATOR_PUSH))
def push_inference_simulator(
        inference_simulator_public_id: Optional[str] = typer.Option(
            None,
            '--inference-simulator-id',
            '-isid',
            help=__("Inference simulator ID"),
            is_eager=False,
        ),
        inference_simulator_slug: Optional[str] = typer.Option(
            None,
            '--inference-simulator-name',
            '-isn',
            help=__("Inference simulator name"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_INFERENCE_SIMULATOR_PUSH
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [inference_simulator_public_id, inference_simulator_slug]) != 1:
        typer.echo("Provide a single identifier for inference simulator - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    inference_simulator_service = service_factory.get_inference_simulator_service()

    inference_simulator_service.project_push_inference_simulator(
        public_id=inference_simulator_public_id,
        slug=inference_simulator_slug,
    )

    raise typer.Exit(0)


@app.command("ls", help=__("List inference simulators"), **get_command_metadata(CliCommand.PROJECT_INFERENCE_SIMULATOR_LS))
def list_inference_simulators(
        project_public_id: Optional[str] = typer.Option(
            None,
            '--project-id',
            '-pid',
            help=__("Project ID. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        project_slug: Optional[str] = typer.Option(
            None,
            '--project-name',
            '-pn',
            help=__("Project name. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
        statuses: List[str] = typer.Option(
            None,
            '--status',
            '-s',
            help=__("Filter by status, use --status all to list all inference simulators"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_INFERENCE_SIMULATOR_LS
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [project_public_id, project_slug]) > 1:
        typer.echo("Provide a single identifier for project - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    inference_simulator_service = service_factory.get_inference_simulator_service()

    inference_simulator_service.print_inference_simulator_list(
        project_public_id=project_public_id,
        project_slug=project_slug,
        statuses=statuses,
        row=row,
        page=page
    )

    typer.echo(__("Inference simulators listing complete"))
    raise typer.Exit(0)

@app.command(name="logs", no_args_is_help=True, help=__("Stream real-time inference simulator logs or view last logs for an inference simulator"), **get_command_metadata(CliCommand.PROJECT_INFERENCE_SIMULATOR_LOGS))
def inference_simulator_logs(
        #TODO doesn't work with public_id
        public_id: Optional[str] = typer.Option(
            None,
            '--inference-simulator-id',
            '-isid',
            help="Inference simulator ID",
            is_eager=False,
        ),
        slug: Optional[str] = typer.Option(
            None,
            '--inference-simulator-name',
            '-isn',
            help="Inference simulator name",
            is_eager=False,
        ),
        logs_number: Optional[int] = typer.Option(
            None,
            '--number',
            '-n',
            help=__("Display a number of latest log entries. No real-time stream if provided."),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_INFERENCE_SIMULATOR_LOGS
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [public_id, slug]) != 1:
        typer.echo("Provide a single identifier for inference simulator - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    logging_service: LoggingService = service_factory.get_logging_service()

    if logs_number is None:
        logging_service.stream_inference_simulator_logs_with_controls(
            public_id=public_id,
            slug=slug,
        )
    else:
        inference_simulator_api_client = service_factory.get_inference_simulator_api_client()

        get_inference_simulator_response: Optional[
            GetInferenceSimulatorResponse] = inference_simulator_api_client.get_inference_simulator(
            public_id=public_id,
            slug=slug,
        )
        if not get_inference_simulator_response:
            typer.echo("Inference simulator not found")
            raise typer.Exit(1)
        else:
            inference_simulator_public_id = get_inference_simulator_response.inferenceSimulator.public_id
            logging_service.print_last_inference_simulator_logs(inference_simulator_public_id=inference_simulator_public_id, logs_number=logs_number)

    app_logger.info(f'Inference simulator log streaming completed')
    raise typer.Exit(0)


from typing import Optional, List

import typer

from thestage.task.communication import task_command
from thestage.cli_command import CliCommand
from thestage.cli_command_helper import get_command_metadata, check_command_permission, get_command_group_help_panel
from thestage.controllers.utils_controller import validate_config_and_get_service_factory, get_current_directory
from thestage.helpers.logger.app_logger import app_logger
from thestage.i18n.translation import __

app = typer.Typer(no_args_is_help=True, help=__("Manage projects"))
config_app = typer.Typer(no_args_is_help=True, help=__("Manage project config"))
app.add_typer(config_app, name="config", rich_help_panel=get_command_group_help_panel())
app.add_typer(task_command.runner_app, name="")


@app.command(name='clone', no_args_is_help=True, help=__("Clone project repository to empty directory"), **get_command_metadata(CliCommand.PROJECT_CLONE))
def clone(
        project_public_id: Optional[str] = typer.Option(
            None,
            "--project-id",
            "-pid",
            help=__("Project ID. ID or name is required"),
            is_eager=False,
        ),
        project_slug: Optional[str] = typer.Option(
            None,
            "--project-name",
            "-pn",
            help=__("Project name. ID or name is required."),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to the working directory: current directory used by default"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_CLONE
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [project_public_id, project_slug]) != 1:
        typer.echo("Provide a single identifier for the project - name or ID.")
        raise typer.Exit(1)

    if not working_directory:
        project_dir_name = project_public_id if project_slug is None else project_slug
        working_directory = get_current_directory().joinpath(project_dir_name)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    project_service = service_factory.get_project_service()

    project_service.clone_project(
        project_slug=project_slug,
        project_public_id=project_public_id,
    )

    raise typer.Exit(0)


@app.command(name='init', no_args_is_help=True, help=__("Initialize project repository with existing files"), **get_command_metadata(CliCommand.PROJECT_INIT))
def init(
        project_public_id: Optional[str] = typer.Option(
            None,
            "--project-id",
            "-pid",
            help=__("Project ID. ID or name is required"),
            is_eager=False,
        ),
        project_slug: Optional[str] = typer.Option(
            None,
            "--project-name",
            "-pn",
            help=__("Project name. ID or name is required"),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
        remote_url: Optional[str] = typer.Option(
            None,
            "--remote",
            "-r",
            help=__("Public Git repository URL to initialize project with"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_INIT
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [project_public_id, project_slug]) != 1:
        typer.echo("Provide a single identifier for the project - name or ID.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    project_service = service_factory.get_project_service()
    project_config = service_factory.get_config_provider().read_project_config()

    if project_config:
        typer.echo(__("Directory is initialized and already contains working project"))
        raise typer.Exit(1)

    project_service.init_project(
        project_slug=project_slug,
        project_public_id=project_public_id,
        remote_url=remote_url,
    )

    raise typer.Exit(0)



@app.command(name='checkout', no_args_is_help=True, help=__("Checkout project repository to a specific reference"), **get_command_metadata(CliCommand.PROJECT_CHECKOUT))
def checkout_project(
        task_public_id: Optional[str] = typer.Option(
            None,
            "--task-id",
            "-tid",
            help="Task ID to checkout",
            is_eager=False,
        ),
        branch_name: Optional[str] = typer.Option(
            None,
            "--branch",
            "-b",
            help="Branch name to checkout. Use '/' value to checkout to the main branch.",
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_CHECKOUT
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [branch_name, task_public_id]) != 1:
        typer.echo("Provide a single identifier for checkout - task ID or branch name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    project_service = service_factory.get_project_service()

    final_branch_name = branch_name
    if branch_name == "/":
        final_branch_name = None

    project_service.checkout_project(
        task_public_id=task_public_id,
        branch_name=final_branch_name,
    )

    raise typer.Exit(0)


@app.command(name='pull', help=__("Pulls the changes from the remote project repository. Equivalent to 'git pull'."), **get_command_metadata(CliCommand.PROJECT_PULL))
def pull_project(
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_PULL
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    project_service = service_factory.get_project_service()

    project_service.pull_project()

    raise typer.Exit(0)


@app.command(name='reset', help=__("Resets the current project branch to remote counterpart. All working tree changes will be lost. Equivalent to 'git fetch && git reset --hard origin/{ref}'."), **get_command_metadata(CliCommand.PROJECT_RESET))
def reset_project(
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_RESET
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    project_service = service_factory.get_project_service()

    project_service.reset_project()

    raise typer.Exit(0)


@config_app.command(name='set-default-container', no_args_is_help=True, help=__("Set default docker container for a project installation"), **get_command_metadata(CliCommand.PROJECT_CONFIG_SET_DEFAULT_CONTAINER))
def set_default_container(
        docker_container_public_id: Optional[str] = typer.Option(
            None,
            '--container-id',
            '-cid',
            help=__("Docker container ID"),
            is_eager=False,
        ),
        docker_container_slug: Optional[str] = typer.Option(
            None,
            '--container-name',
            '-cn',
            help=__("Docker container name"),
            is_eager=False,
        ),
        unset_default_container: Optional[bool] = typer.Option(
            False,
            "--unset",
            "-u",
            help=__("Unsets the default docker container"),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_CONFIG_SET_DEFAULT_CONTAINER
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    container_args_count = sum(v is not None for v in [docker_container_public_id, docker_container_slug])
    if container_args_count > 1:
        typer.echo("Provide a single identifier for the container - name or ID.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)

    if unset_default_container and container_args_count > 0:
        typer.echo("Cannot provide container identifier and '--unset' flag simultaneously")
        raise typer.Exit(1)

    if not unset_default_container and container_args_count != 1:
        typer.echo("Provide container identifier or use '--unset' flag")
        raise typer.Exit(1)

    project_service = service_factory.get_project_service()

    project_service.set_default_container(
        container_public_id=docker_container_public_id,
        container_slug=docker_container_slug,
    )

    raise typer.Exit(0)


@config_app.command(name='get', no_args_is_help=False, help=__("View config for a local project installation"), **get_command_metadata(CliCommand.PROJECT_CONFIG_GET))
def get_project_config(
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_CONFIG_GET
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    project_service = service_factory.get_project_service()

    project_service.print_project_config()

    raise typer.Exit(0)


@app.command(name='attach-instance', no_args_is_help=True, help=__("Attach instance to project"), **get_command_metadata(CliCommand.PROJECT_ATTACH_INSTANCE))
def attach_instance(
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
        project_public_id: Optional[str] = typer.Option(
            None,
            '--project-id',
            '-pid',
            help=__("Project ID. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        project_slug: Optional[str] = typer.Option(
            None,
            '--project-name',
            '-pn',
            help=__("Project name. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        rented_instance_public_id: Optional[List[str]] = typer.Option(
            None,
            '--rented-instance-id',
            '-rid',
            help=__("Rented instance ID which will be attached to project. Can be used multiple times."),
            is_eager=False,
        ),
        rented_instance_slug: Optional[List[str]] = typer.Option(
            None,
            '--rented-instance-name',
            '-rn',
            help=__("Rented instance name which will be attached to project. Can be used multiple times."),
            is_eager=False,
        ),
        self_hosted_instance_public_id: Optional[List[str]] = typer.Option(
            None,
            '--self-hosted-instance-id',
            '-sid',
            help=__("Self-hosted instance ID which will be attached to project. Can be used multiple times."),
            is_eager=False,
        ),
        self_hosted_instance_slug: Optional[List[str]] = typer.Option(
            None,
            '--self-hosted-instance-name',
            '-sn',
            help=__("Self-hosted instance name which will be attached to project. Can be used multiple times."),
            is_eager=False,
        ),

):
    command_name = CliCommand.PROJECT_ATTACH_INSTANCE
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [project_public_id, project_slug]) > 1:
        typer.echo("Provide a single identifier for project - ID or name.")
        raise typer.Exit(1)

    if sum(v is not None for v in [rented_instance_public_id, rented_instance_slug, self_hosted_instance_public_id,
                                   self_hosted_instance_slug]) != 1:
        typer.echo("Provide exactly one identifier (ID or Name) for either a rented or self-hosted instance.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    project_service = service_factory.get_project_service()

    project_service.modify_project_instance_assignment(
        project_public_id=project_public_id,
        project_slug=project_slug,
        rented_instance_public_id_to_assign=rented_instance_public_id,
        rented_instance_slug_to_assign=rented_instance_slug,
        self_hosted_instance_public_id_to_assign=self_hosted_instance_public_id,
        self_hosted_instance_slug_to_assign=self_hosted_instance_slug,
    )

    raise typer.Exit(0)


@app.command(name='detach-instance', no_args_is_help=True, help=__("Detach instance from project"), **get_command_metadata(CliCommand.PROJECT_DETACH_INSTANCE))
def detach_instance(
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
        project_public_id: Optional[str] = typer.Option(
            None,
            '--project-id',
            '-pid',
            help=__("Project ID. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        project_slug: Optional[str] = typer.Option(
            None,
            '--project-name',
            '-pn',
            help=__("Project name. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        rented_instance_public_id: Optional[List[str]] = typer.Option(
            None,
            '--rented-instance-id',
            '-rid',
            help=__("Rented instance ID which will be detached from project. Can be used multiple times."),
            is_eager=False,
        ),
        rented_instance_slug: Optional[List[str]] = typer.Option(
            None,
            '--rented-instance-name',
            '-rn',
            help=__("Rented instance name which will be detached from project. Can be used multiple times."),
            is_eager=False,
        ),
        self_hosted_instance_public_id: Optional[List[str]] = typer.Option(
            None,
            '--self-hosted-instance-id',
            '-sid',
            help=__("Self-hosted instance ID which will be detached from project. Can be used multiple times."),
            is_eager=False,
        ),
        self_hosted_instance_slug: Optional[List[str]] = typer.Option(
            None,
            '--self-hosted-instance-name',
            '-sn',
            help=__("Self-hosted instance name which will be detached from project. Can be used multiple times."),
            is_eager=False,
        ),

):
    command_name = CliCommand.PROJECT_DETACH_INSTANCE
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [project_public_id, project_slug]) > 1:
        typer.echo("Provide a single identifier for project - ID or name.")
        raise typer.Exit(1)

    if sum(v is not None for v in [rented_instance_public_id, rented_instance_slug, self_hosted_instance_public_id,
                                   self_hosted_instance_slug]) != 1:
        typer.echo("Provide exactly one identifier (ID or Name) for either a rented or self-hosted instance.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    project_service = service_factory.get_project_service()

    project_service.modify_project_instance_assignment(
        project_public_id=project_public_id,
        project_slug=project_slug,
        rented_instance_public_id_to_delete=rented_instance_public_id,
        rented_instance_slug_to_delete=rented_instance_slug,
        self_hosted_instance_public_id_to_delete=self_hosted_instance_public_id,
        self_hosted_instance_slug_to_delete=self_hosted_instance_slug,
    )

    raise typer.Exit(0)


@app.command(name='delete', help=__("Delete project"), **get_command_metadata(CliCommand.PROJECT_DELETE))
def delete_project(
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            is_eager=False,
        ),
        project_public_id: Optional[str] = typer.Option(
            None,
            '--project-id',
            '-pid',
            help=__("Project ID. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        project_slug: Optional[str] = typer.Option(
            None,
            '--project-name',
            '-pn',
            help=__("Project name. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_DELETE
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)
    if sum(v is not None for v in [project_public_id, project_slug]) > 1:
        typer.echo("Provide a single identifier for project - ID or name.")
        raise typer.Exit(1)
    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    project_service = service_factory.get_project_service()

    project_service.delete_project(
        project_public_id=project_public_id,
        project_slug=project_slug,
    )

    raise typer.Exit(0)


@app.command(name='ls', help=__("List projects. Can be filtered by instances."), **get_command_metadata(CliCommand.PROJECT_LS))
def project_list(
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
        rented_instance_public_id: Optional[str] = typer.Option(
            None,
            '--rented-instance-id',
            '-rid',
            help=__("Filter projects by rented instance ID"),
            is_eager=False,
        ),
        rented_instance_slug: Optional[str] = typer.Option(
            None,
            '--rented-instance-name',
            '-rn',
            help=__("Filter projects by rented instance name"),
            is_eager=False,
        ),
        self_hosted_instance_public_id: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-id',
            '-sid',
            help=__("Filter projects by self-hosted instance ID"),
            is_eager=False,
        ),
        self_hosted_instance_slug: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-name',
            '-sn',
            help=__("Filter projects by self-hosted instance name"),
            is_eager=False,
        ),
        statuses: List[str] = typer.Option(
            None,
            '--status',
            '-s',
            help=__("Filter by status, use --status all to list all rented server instances"),
            is_eager=False,
        ),

):
    command_name = CliCommand.PROJECT_LS
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    service_factory = validate_config_and_get_service_factory()
    project_service = service_factory.get_project_service()

    project_service.print_project_list(
        row=row,
        page=page,
        statuses=statuses,
        rented_instance_public_id=rented_instance_public_id,
        rented_instance_slug=rented_instance_slug,
        self_hosted_instance_public_id=self_hosted_instance_public_id,
        self_hosted_instance_slug=self_hosted_instance_slug,
    )

    raise typer.Exit(0)


from enum import Enum

# TODO reduce usage of statuses to zero
class InferenceModelStatus(str, Enum):
    SCHEDULED: str = 'SCHEDULED'
    PROCESSING: str = 'PROCESSING'
    PUSH_SUCCEED: str = 'PUSH_SUCCEED'
    PUSH_FAILED: str = 'PUSH_FAILED'
    UNKNOWN: str = 'UNKNOWN'
    ALL: str = 'ALL'


from thestage import __app_name__
from thestage.controllers.base_controller import app as main_app

from thestage.docker_container.communication.docker_command import app as container_app
from thestage.instance.communication.instance_command import app as instance_app
from thestage.project.communication.project_command import app as project_app
from thestage.config.communication.config_command import app as config_app
from thestage.task.communication.task_command import app as task_app

main_app.add_typer(container_app, name="container")
main_app.add_typer(instance_app, name="instance")
main_app.add_typer(config_app, name="config")
main_app.add_typer(project_app, name="project")
main_app.add_typer(task_app, name="task")

def main():
    project_app([
        "run",
        "-wd",
        "/Users/alexey/Documents/clonetest",
        "-cid",
        '62-1',
        '-t',
        'task',
        "echo 123",
        "-nl"
    ], prog_name=__app_name__)

if __name__ == "__main__":
    main()


from typing import Optional, Dict

from pydantic import Field, BaseModel, ConfigDict

from thestage.services.clients.thestage_api.dtos.frontend_status import FrontendStatusDto
from thestage.services.clients.thestage_api.dtos.base_response import TheStageBasePaginatedResponse
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList
from thestage.instance.dto.enum.gpu_name import InstanceGpuType
from thestage.instance.dto.enum.cpu_type import InstanceCpuType


class InstanceRentedDto(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    public_id: Optional[str] = Field(None, alias='publicId')
    slug: Optional[str] = Field(None, alias='slug')
    cpu_type: Optional[InstanceCpuType] = Field(InstanceCpuType.UNKNOWN, alias='cpuType')
    cpu_cores: Optional[int] = Field(None, alias='cpuCores')
    gpu_type: Optional[InstanceGpuType] = Field(InstanceGpuType.UNKNOWN, alias='gpuType')
    frontend_status: Optional[FrontendStatusDto] = Field(None, alias='frontendStatus')
    ip_address: Optional[str] = Field(None, alias='ipAddress')
    host_username: Optional[str] = Field(None, alias='hostUsername')


class InstanceRentedListResponse(TheStageBasePaginatedResponse):
    paginated_list: Optional[PaginatedEntityList[InstanceRentedDto]] = Field(None, alias='paginatedEntityList')


class InstanceRentedItemResponse(TheStageBasePaginatedResponse):
    model_config = ConfigDict(use_enum_values=True)

    instance_rented: Optional[InstanceRentedDto] = Field(None, alias='instanceRented')


class InstanceRentedBusinessStatusMapperResponse(TheStageBasePaginatedResponse):
    model_config = ConfigDict(use_enum_values=True)

    instance_rented_business_status_map: Dict[str, str] = Field(default={}, alias='instanceRentedBusinessStatusMap')



from typing import Optional

from pydantic import BaseModel, ConfigDict, Field

from thestage.services.clients.thestage_api.dtos.frontend_status import FrontendStatusDto


class Task(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    public_id: Optional[str] = Field(None, alias='publicId')
    docker_container_public_id: Optional[str] = Field(None, alias='dockerContainerPublicId')
    title: Optional[str] = Field(None, alias='title')
    frontend_status: FrontendStatusDto = Field(None, alias='frontendStatus')
    commit_hash: Optional[str] = Field(None, alias='commitHash')
    started_at: Optional[str] = Field(None, alias='startedAt')
    finished_at: Optional[str] = Field(None, alias='finishedAt')



from typing import Optional

from pydantic import Field

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.task.dto.task import Task


class TaskViewResponse(TheStageBaseResponse):
    task: Optional[Task] = Field(None, alias='task')



from thestage.exceptions.base_exception import BaseAbstractException


class LogPollingException(BaseAbstractException):
    def __init__(self, message: str):
        super(LogPollingException, self).__init__(message=message)



from typing import Optional

from pydantic import Field, ConfigDict, BaseModel


class RunTaskRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    projectPublicId: str = Field(None, alias='projectPublicId')
    dockerContainerPublicId: str = Field(None, alias='dockerContainerPublicId')
    commitHash: Optional[str] = Field(None, alias='commitHash')
    runCommand: str = Field(None, alias='runCommand')
    taskTitle: Optional[str] = Field(None, alias='taskTitle')



from typing import Optional, List

from pydantic import Field, ConfigDict

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.logging.dto.log_message import LogMessage


class UserLogsQueryResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)

    queryResult: Optional[List[LogMessage]] = Field(None, alias='queryResult')
    entriesFound: Optional[int] = Field(None, alias='entriesFound')
    totalCount: Optional[int] = Field(None, alias='totalCount')



from thestage.exceptions.base_exception import BaseAbstractException


class GitAccessException(BaseAbstractException):

    __url: str

    def __init__(self, message: str, url: str, dop_message: str):
        super(GitAccessException, self).__init__(message=message)
        self.__url = url
        self.__dop_message = dop_message

    def get_url(self) -> str:
        return self.__url

    def get_dop_message(self) -> str:
        return self.__dop_message



from typing import Optional

from pydantic import Field, ConfigDict, BaseModel


class StartInferenceSimulatorRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    projectPublicId: str = Field(None, alias='projectPublicId')
    instanceRentedPublicId: Optional[str] = Field(None, alias='instanceRentedPublicId')
    instanceRentedSlug: Optional[str] = Field(None, alias='instanceRentedSlug')
    selfhostedInstancePublicId: Optional[str] = Field(None, alias='selfhostedInstancePublicId')
    selfhostedInstanceSlug: Optional[str] = Field(None, alias='selfhostedInstanceSlug')
    commitHash: Optional[str] = Field(None, alias='commitHash')
    inferenceDir: Optional[str] = Field(None, alias='inferenceDir')
    isSkipInstallation: Optional[bool] = Field(False, alias='isSkipInstallation')


from typing import Optional, List, Dict

from thestage.config.business.config_provider import ConfigProvider
from thestage.global_dto.enums.order_direction_type import OrderDirectionType
from thestage.helpers.error_handler import error_handler
from thestage.services.clients.thestage_api.core.api_client_core import TheStageApiClientCore
from thestage.services.clients.thestage_api.dtos.entity_filter_request import EntityFilterRequest
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList

from thestage.inference_model.dto.inference_model import InferenceModel
from thestage.inference_model.dto.inference_simulator_model_list_for_project_request import InferenceSimulatorModelListForProjectRequest
from thestage.inference_model.dto.inference_simulator_model_list_for_project_response import InferenceSimulatorModelListForProjectResponse
from thestage.inference_model.dto.push_inference_simulator_model_request import PushInferenceSimulatorModelRequest
from thestage.inference_model.dto.push_inference_simulator_model_response import PushInferenceSimulatorModelResponse
from thestage.inference_model.dto.inference_simulator_model_response import InferenceSimulatorModelStatusMapperResponse
from thestage.inference_model.dto.deploy_inference_model_to_instance_request import DeployInferenceModelToInstanceRequest
from thestage.inference_model.dto.deploy_inference_model_to_instance_response import DeployInferenceModelToInstanceResponse
from thestage.inference_model.dto.deploy_inference_model_to_sagemaker_request import DeployInferenceModelToSagemakerRequest
from thestage.inference_model.dto.deploy_inference_model_to_sagemaker_response import DeployInferenceModelToSagemakerResponse


class InferenceModelApiClient(TheStageApiClientCore):
    def __init__(self, config_provider: ConfigProvider):
        super().__init__(url=config_provider.get_config().main.thestage_api_url)
        self.__config_provider = config_provider

    def get_inference_simulator_model_list_for_project(
            self,
            project_public_id: Optional[str],
            project_slug: Optional[str],
            statuses: Optional[List[str]] = None,
            page: int = 1,
            limit: int = 10,
    ) -> Optional[PaginatedEntityList[InferenceModel]]:
        request = InferenceSimulatorModelListForProjectRequest(
            projectPublicId=project_public_id,
            projectSlug=project_slug,
            statuses=statuses,
            entityFilterRequest=EntityFilterRequest(
                orderByField="createdAt",
                orderByDirection=OrderDirectionType.DESC,
                page=page,
                limit=limit,
            ),
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/inference-simulator-model/list',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )
        result = InferenceSimulatorModelListForProjectResponse.model_validate(response) if response else None
        return result.inferenceSimulatorModels if result and result.isSuccess else None


    def push_project_inference_simulator_model(
            self,
            public_id: Optional[str],
            slug: Optional[str],
    ) -> Optional[PushInferenceSimulatorModelResponse]:
        request = PushInferenceSimulatorModelRequest(
            inferenceSimulatorPublicId=public_id,
            inferenceSimulatorSlug=slug,
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/inference-simulator/push-model',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        return PushInferenceSimulatorModelResponse.model_validate(response) if response else None


    def get_inference_simulator_model_business_status_map(self) -> Optional[Dict[str, str]]:
        response = self._request(
            method='POST',
            url='/user-api/v1/inference-simulator-model/status-localized-mapping',
            data=None,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        data = InferenceSimulatorModelStatusMapperResponse.model_validate(response) if response else None

        return data.inference_simulator_model_status_map if data else None


    @error_handler()
    def deploy_inference_model_to_instance(
            self,
            model_public_id: str,
            model_slug: str,
            new_inference_simulator_slug: str,
            rented_instance_public_id: Optional[str] = None,
            rented_instance_slug: Optional[str] = None,
            self_hosted_instance_public_id: Optional[str] = None,
            self_hosted_instance_slug: Optional[str] = None,

    ) -> Optional[DeployInferenceModelToInstanceResponse]:
        request = DeployInferenceModelToInstanceRequest(
            modelPublicId=model_public_id,
            modelSlug=model_slug,
            instanceRentedPublicId=rented_instance_public_id,
            instanceRentedSlug=rented_instance_slug,
            selfhostedInstancePublicId=self_hosted_instance_public_id,
            selfhostedInstanceSlug=self_hosted_instance_slug
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/inference-simulator-model/deploy/instance',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )
        return DeployInferenceModelToInstanceResponse.model_validate(response) if response else None


    @error_handler()
    def deploy_inference_model_to_sagemaker(
            self,
            model_public_id: Optional[str],
            model_slug: Optional[str],
            arn: Optional[str] = None,
    ) -> Optional[DeployInferenceModelToSagemakerResponse]:
        request = DeployInferenceModelToSagemakerRequest(
            modelPublicId=model_public_id,
            modelSlug=model_slug,
            arn=arn,
        )

        response = self._request(
            method='POST',
            url='/user-api/v1/inference-simulator-model/grant-user-arn-access',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )
        return DeployInferenceModelToSagemakerResponse.model_validate(response) if response else None


from typing import Optional

from thestage.task.dto.task_entity import TaskEntity
from thestage.services.abstract_mapper import AbstractMapper
from thestage.task.dto.task import Task


class TaskMapper(AbstractMapper):

    def build_entity(self, item: Task) -> Optional[TaskEntity]:
        if not item:
            return None

        return TaskEntity(
            public_id=item.public_id or '',
            title=item.title or '',
            status=item.frontend_status.status_translation or '',
            docker_container_public_id=item.docker_container_public_id,
            started_at=item.started_at or '',
            finished_at=item.finished_at or '',
        )



from typing import List, Optional, Any, Generic, TypeVar

from pydantic import Field, ConfigDict, BaseModel

from thestage.services.clients.thestage_api.dtos.pagination_data import PaginationData

T = TypeVar('T')

class PaginatedEntityList(BaseModel, Generic[T]):
    entities: List[T] = Field(default_factory=list, alias='entities')
    pagination_data: Optional[PaginationData] = Field(None, alias='paginationData')



from typing import Optional

from pydantic import Field, ConfigDict

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse


class AddSshKeyToUserResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)

    sshKeyPairPublicId: Optional[str] = Field(None, alias='sshKeyPairPublicId')


from pydantic import ConfigDict

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse


class AddSshPublicKeyToInstanceResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)




import time
from datetime import datetime
from typing import Optional, List

import boto3
import typer

from thestage.config.business.config_provider import ConfigProvider
from thestage.helpers.error_handler import error_handler
from thestage.i18n.translation import __
from thestage.inference_model.business.mapper.inference_model_mapper import InferenceModelMapper
from thestage.inference_model.communication.inference_model_api_client import InferenceModelApiClient
from thestage.inference_model.dto.deploy_inference_model_to_instance_response import \
    DeployInferenceModelToInstanceResponse
from thestage.inference_model.dto.deploy_inference_model_to_sagemaker_response import \
    DeployInferenceModelToSagemakerResponse
from thestage.inference_model.dto.enum.inference_model_status import InferenceModelStatus
from thestage.inference_model.dto.inference_model import InferenceModel
from thestage.inference_model.dto.inference_model_entity import InferenceModelEntity
from thestage.project.business.project_service import ProjectService
from thestage.project.dto.project_config import ProjectConfig
from thestage.services.abstract_service import AbstractService
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList


class InferenceModelService(AbstractService):
    def __init__(
            self,
            inference_model_api_client: InferenceModelApiClient,
            config_provider: ConfigProvider,
            project_service: ProjectService,
    ):
        self.__inference_model_api_client = inference_model_api_client
        self.__config_provider = config_provider
        self.__project_service = project_service


    @error_handler()
    def get_project_inference_simulator_model_list(
            self,
            project_public_id: Optional[str],
            project_slug: Optional[str],
            statuses: List[str],
            row: int = 5,
            page: int = 1,
    ) -> PaginatedEntityList[InferenceModel]:
        data: Optional[
            PaginatedEntityList[InferenceModel]] = self.__inference_model_api_client.get_inference_simulator_model_list_for_project(
            statuses=statuses,
            project_public_id=project_public_id,
            project_slug=project_slug,
            page=page,
            limit=row,
        )

        return data

    @error_handler()
    def project_deploy_inference_simulator_model_to_instance(
            self,
            model_public_id: Optional[str] = None,
            model_slug: Optional[str] = None,
            rented_instance_public_id: Optional[str] = None,
            rented_instance_slug: Optional[str] = None,
            self_hosted_instance_public_id: Optional[str] = None,
            self_hosted_instance_slug: Optional[str] = None,
    ) -> str:
        config = self.__config_provider.get_config()
        project_config: ProjectConfig = self.__project_service.get_fixed_project_config(do_check_default_container=False)
        if not project_config:
            typer.echo(
                __("No project found at the path: %path%. Initialize or clone a project first. Or provide path to project using --working-directory option.",
                   {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        instance_args_count = sum(v is not None for v in
                                  [rented_instance_public_id, rented_instance_slug, self_hosted_instance_public_id,
                                   self_hosted_instance_slug])
        if instance_args_count != 1:
            typer.echo("Provide a single instance (rented or self-hosted) identifier - name or ID.")
            raise typer.Exit(1)

        model_args_count = sum(v is not None for v in [model_public_id, model_slug])
        if model_args_count != 1:
            typer.echo("Provide a single model identifier - name or ID.")
            raise typer.Exit(1)

        typer.echo(f"Creating inference simulator")
        deploy_model_to_instance_response: DeployInferenceModelToInstanceResponse = self.__inference_model_api_client.deploy_inference_model_to_instance(
            model_public_id=model_public_id,
            model_slug=model_slug,
            rented_instance_public_id=rented_instance_public_id,
            rented_instance_slug=rented_instance_slug,
            self_hosted_instance_public_id=self_hosted_instance_public_id,
            self_hosted_instance_slug=self_hosted_instance_slug,
        )
        if deploy_model_to_instance_response:
            if deploy_model_to_instance_response.message:
                typer.echo(deploy_model_to_instance_response.message)
            if deploy_model_to_instance_response.isSuccess:
                typer.echo("Inference simulator has been scheduled to run successfully.")
            else:
                typer.echo(__(
                    'Failed to start inference simulator: %server_massage%',
                    {'server_massage': deploy_model_to_instance_response.message or ""}
                ))
                raise typer.Exit(1)
        else:
            typer.echo(__("Failed to start inference simulator"))
            raise typer.Exit(1)

        return deploy_model_to_instance_response.inferenceSimulatorPublicId


    @error_handler()
    def project_deploy_inference_simulator_model_to_sagemaker(
            self,
            model_public_id: Optional[str] = None,
            model_slug: Optional[str] = None,
            arn: Optional[str] = None,
            instance_type: Optional[str] = None,
            initial_variant_weight: Optional[float] = 1.0,
            initial_instance_count: Optional[int] = None,
    ) -> None:
        config = self.__config_provider.get_config()
        project_config: ProjectConfig = self.__project_service.get_fixed_project_config(do_check_default_container=False)
        if not project_config:
            typer.echo(
                __("No project found at the path: %path%. Initialize or clone a project first. Or provide path to project using --working-directory option.",
                   {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        if not instance_type:
            typer.echo(__("Error: Instance type is required."))
            raise typer.Exit(1)

        if not initial_instance_count:
            typer.echo(__("Error: Initial instance count is required."))
            raise typer.Exit(1)

        if not arn:
            typer.echo(__("Error: ARN is required."))
            raise typer.Exit(1)

        project_config: ProjectConfig = self.__config_provider.read_project_config()
        if not project_config:
            typer.echo(__("No project found at the path: %path%. Initialize or clone a project first.",
                          {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        deploy_model_to_sagemaker_response: DeployInferenceModelToSagemakerResponse = self.__inference_model_api_client.deploy_inference_model_to_sagemaker(
            model_public_id=model_public_id,
            model_slug=model_slug,
            arn=arn,
        )

        if not deploy_model_to_sagemaker_response.isSuccess:
            typer.echo(__(
                'Model deployment preparation failed: %server_massage%',
                {'server_massage': deploy_model_to_sagemaker_response.message or ""}
            ))
            raise typer.Exit(1)

        model_id = deploy_model_to_sagemaker_response.modelId
        image_uri = deploy_model_to_sagemaker_response.ecrImageUrl
        model_uri = deploy_model_to_sagemaker_response.s3ArtifactsUrl
        region = "us-east-1"
        sm_client = boto3.client('sagemaker', region_name=region)

        try:
            container = {
                "Image": image_uri,
                "ModelDataUrl": model_uri,
                "Environment": {
                    "SAGEMAKER_TRITON_DEFAULT_MODEL_NAME": model_id,
                    "THESTAGE_API_URL": config.main.thestage_api_url,
                    "THESTAGE_AUTH_TOKEN": config.main.thestage_auth_token
                },
            }

            sm_model_name = f"{model_slug}-{datetime.now().strftime('%Y%m%d-%H%M%S')}"
            create_model_response = sm_client.create_model(
                ModelName=sm_model_name,
                ExecutionRoleArn=arn,
                PrimaryContainer=container,
            )
            typer.echo(f"Model created successfully. Model ARN: {create_model_response['ModelArn']}")

            endpoint_config_name = f"{model_slug}-{datetime.now().strftime('%Y%m%d-%H%M%S')}"
            create_endpoint_config_response = sm_client.create_endpoint_config(
                EndpointConfigName=endpoint_config_name,
                ProductionVariants=[
                    {
                        "InstanceType": instance_type,
                        "InitialVariantWeight": initial_variant_weight,
                        "InitialInstanceCount": initial_instance_count,
                        "ModelName": sm_model_name,
                        "VariantName": "AllTraffic",
                    }
                ],
            )
            typer.echo(
                f"Endpoint configuration created successfully. Endpoint Config ARN: {create_endpoint_config_response['EndpointConfigArn']}")

            endpoint_name = f"{model_slug}-{datetime.now().strftime('%Y%m%d-%H%M%S')}"
            create_endpoint_response = sm_client.create_endpoint(
                EndpointName=endpoint_name,
                EndpointConfigName=endpoint_config_name,
            )
            typer.echo(f"Endpoint created successfully. Endpoint ARN: {create_endpoint_response['EndpointArn']}")

            typer.echo("Waiting for the endpoint to become active...")
            while True:
                resp = sm_client.describe_endpoint(EndpointName=endpoint_name)
                status = resp["EndpointStatus"]
                typer.echo(f"Status: {status}")
                if status == "InService":
                    break
                elif status == "Failed":
                    typer.echo(f"Endpoint creation failed. Reason: {resp.get('FailureReason', 'Unknown')}")
                    raise typer.Exit(1)
                time.sleep(60)

            typer.echo(f"Endpoint is ready. ARN: {resp['EndpointArn']} Status: {status}")

        except Exception as e:
            typer.echo(__("Failed to deploy the inference simulator model to SageMaker: %error%", {"error": str(e)}))
            raise typer.Exit(1)


    @error_handler()
    def print_inference_simulator_model_list(self, project_public_id, project_slug, statuses, row, page):
        if not project_public_id and not project_slug:
            project_config: ProjectConfig = self.__config_provider.read_project_config()
            if not project_config:
                typer.echo(__("Provide the project unique ID or name, or run this command from within an initialized project directory"))
                raise typer.Exit(1)
            project_public_id = project_config.public_id

        inference_simulator_model_status_map = self.__inference_model_api_client.get_inference_simulator_model_business_status_map()

        if not statuses:
            statuses = ({key: inference_simulator_model_status_map[key] for key in [
                InferenceModelStatus.SCHEDULED,
                InferenceModelStatus.PROCESSING,
                InferenceModelStatus.PUSH_SUCCEED,
            ]}).values()

        if "all" in statuses:
            statuses = inference_simulator_model_status_map.values()

        for input_status_item in statuses:
            if input_status_item not in inference_simulator_model_status_map.values():
                typer.echo(__("'%invalid_status%' is not one of %valid_statuses%", {
                    'invalid_status': input_status_item,
                    'valid_statuses': str(list(inference_simulator_model_status_map.values()))
                }))
                raise typer.Exit(1)

        typer.echo(__(
            "Listing inference simulator models with the following statuses: %statuses%, to view all inference simulator models, use --status all",
            placeholders={
                'statuses': ', '.join([status_item for status_item in statuses])
            }))

        backend_statuses: List[str] = [key for key, value in inference_simulator_model_status_map.items() if value in statuses]

        self.print(
            func_get_data=self.get_project_inference_simulator_model_list,
            func_special_params={
                'project_public_id': project_public_id,
                'project_slug': project_slug,
                'statuses': backend_statuses,
            },
            mapper=InferenceModelMapper(),
            headers=list(map(lambda x: x.alias, InferenceModelEntity.model_fields.values())),
            row=row,
            page=page,
            max_col_width=[100, 100, 100, 100, 25],
            show_index="never",
        )


import datetime

import os

from pathlib import Path
from typing import Optional, List

import git
import typer
from git import Remote, Repo, GitCommandError, Commit
from gitdb.exc import BadName
from rich import print

from thestage.color_scheme.color_scheme import ColorScheme
from thestage.config.env_base import THESTAGE_CONFIG_DIR
from thestage.exceptions.git_access_exception import GitAccessException
from thestage.git.business.ProgressPrinter import ProgressPrinter
from thestage.services.filesystem_service import FileSystemService


class GitLocalClient:
    __base_name_remote: str = 'origin'
    __base_name_local: str = 'main'
    __git_ignore_thestage_line: str = f'/{THESTAGE_CONFIG_DIR}/'

    __special_main_branches = ['main', 'master']

    __base_git_url: str = 'https://github.com/'

    def __init__(
            self,
            file_system_service: FileSystemService,
    ):
        self.__file_system_service = file_system_service

    # todo delete this
    def __get_repo(self, path: str) -> Repo:
        return git.Repo(path)


    def is_present_local_git(self, path: str) -> bool:
        git_path = self.__file_system_service.get_path(path)
        if not git_path.exists():
            return False

        git_path = git_path.joinpath('.git')
        if not git_path.exists():
            return False

        result = git.repo.base.is_git_dir(git_path)
        return result


    def get_remote(self, path: str) -> Optional[List[Remote]]:
        is_git_repo = self.is_present_local_git(path=path)
        if is_git_repo:
            repo = git.Repo(path)
            remotes: Optional[List[Remote]] = list(repo.remotes) if repo.remotes else []
            return remotes
        return None


    def has_remote(self, path: str) -> bool:
        remotes: Optional[List[Remote]] = self.get_remote(path)
        return True if remotes is not None and len(remotes) > 0 else False


    def has_changes_with_untracked(self, path: str) -> bool:
        repo = self.__get_repo(path=path)
        return repo.is_dirty(untracked_files=True)


    def init_repository(
            self,
            path: str,
    ) -> Optional[Repo]:

        repo = git.Repo.init(path)

        # if repo:
        #     #default git name master, rename to main - sync wih github
        #     repo.git.branch("-M", self.__base_name_local)

        return repo


    def add_remote_to_repo(
            self,
            path: str,
            remote_url: str,
            remote_name: str,
    ) -> bool:
        repo = self.__get_repo(path=path)
        remotes: List[Remote] = repo.remotes
        not_present = True
        if remotes:
            item = list(filter(lambda x: x.name == remote_name, remotes))
            if len(item) > 0:
                not_present = False

        if not_present:
            remote: Remote = repo.create_remote(
                name=self.__base_name_remote,
                url=remote_url,
            )
            if remote:
                return True
            else:
                return False
        else:
            return True


    def git_fetch(self, path: str, deploy_key_path: str):
        repo = self.__get_repo(path=path)
        git_ssh_cmd = 'ssh -F /dev/null -o StrictHostKeyChecking=no -o IdentitiesOnly=yes -i %s' % deploy_key_path

        with repo.git.custom_environment(GIT_SSH_COMMAND=git_ssh_cmd):
            remote: Remote = repo.remote(self.__base_name_remote)
            if remote:
                progress = ProgressPrinter()
                try:
                    remote.fetch(progress=progress)
                except GitCommandError as ex:
                    for line in progress.allDroppedLines():
                        # returning the whole output if failed - so that user have any idea what's going on
                        print(f'>> {line}')
                    raise ex


    def git_pull(self, path: str, deploy_key_path: str):
        repo = self.__get_repo(path=path)
        git_ssh_cmd = 'ssh -F /dev/null -o StrictHostKeyChecking=no -o IdentitiesOnly=yes -i %s' % deploy_key_path

        with repo.git.custom_environment(GIT_SSH_COMMAND=git_ssh_cmd):
            local_branch = self.__base_name_local
            if repo.active_branch.name:
                local_branch = repo.active_branch.name

            origin = repo.remote(self.__base_name_remote)

            if origin:
                repo.git.pull('origin', local_branch)
                typer.echo(f"Pulled remote changes to branch '{local_branch}'")


    def find_main_branch_name(self, path: str) -> Optional[str]:
        repo = self.__get_repo(path=path)
        if repo:
            for branch in [head.name for head in repo.heads]:
                if branch in self.__special_main_branches:
                    return branch
        return None


    def get_active_branch_name(self, path: str) -> Optional[str]:
        repo = self.__get_repo(path=path)
        if repo:
            if repo.head.is_detached:
                return None
            return repo.active_branch.name
        return None


    def git_checkout_to_branch(self, path: str, branch: str):
        repo = self.__get_repo(path=path)
        if repo:
            repo.git.checkout(branch.strip())


    def git_checkout_to_commit(self, path: str, commit_hash: str = None) -> bool:
        repo = self.__get_repo(path=path)
        if repo:
            if self.is_commit_exists(repo, commit_hash):
                repo.git.checkout(commit_hash.strip())
                return True
            else:
                typer.echo(f"Could not checkout to commit {commit_hash} - reference not found in repository")
        return False


    def build_http_repo_url(self, git_path: str) -> str:
        start_path_pos = git_path.find(":")
        pre_url = git_path[start_path_pos + 1:]
        url = pre_url.replace('.git', '')
        return self.__base_git_url + url


    def clone(self, url: str, path: str, deploy_key_path: Optional[str] = None) -> Optional[Repo]:
        env = {}

        if deploy_key_path:
            git_ssh_cmd = 'ssh -F /dev/null -o StrictHostKeyChecking=no -o IdentitiesOnly=yes -i %s' % deploy_key_path
            env = {"GIT_SSH_COMMAND": git_ssh_cmd}

        try:
            return Repo.clone_from(url=url, to_path=path, env=env)
        except GitCommandError as base_ex:
            msg = base_ex.stderr
            if msg and 'Repository not found' in msg and 'correct access rights' in msg:
                raise GitAccessException(
                    message='You dont have access to repository, or repository not found.',
                    url=self.build_http_repo_url(git_path=url),
                    dop_message=msg,
                )
            else:
                raise base_ex


    def commit_local_changes(
            self,
            path: str,
            name: Optional[str] = None,
    ) -> Optional[str]:
        repo = self.__get_repo(path=path)
        if repo.head.is_detached:
            line_color = ColorScheme.GIT_HEADLESS.value
            print(f'[{line_color}]Committing in detached head state at {repo.head.commit.hexsha}[/{line_color}]')
        commit_name = name if name else f"Auto commit {str(datetime.datetime.now().date())}"
        commit = repo.git.commit('--allow-empty', '-m', commit_name, )
        return commit


    def push_changes(
            self,
            path: str,
            deploy_key_path: str
    ):
        repo = self.__get_repo(path=path)
        git_ssh_cmd = 'ssh -F /dev/null -o StrictHostKeyChecking=no -o IdentitiesOnly=yes -i %s' % deploy_key_path

        with repo.git.custom_environment(GIT_SSH_COMMAND=git_ssh_cmd):
            origin = repo.remote(self.__base_name_remote)
            if origin:
                progress = ProgressPrinter()
                # repo.git.push(origin.name, repo.active_branch.name)
                try:
                    origin.push(refspec=repo.active_branch.name, progress=progress).raise_if_error()
                except GitCommandError as ex:
                    for line in progress.allDroppedLines():
                        # returning the whole output if failed - so that user have any idea what's going on
                        print(f'>> {line}')
                    raise ex


    def get_current_commit(self, path: str,) -> Optional[Commit]:
        repo = self.__get_repo(path=path)
        if repo:
            return repo.head.commit
        else:
            return None


    def get_commit_by_hash(self, path: str, commit_hash: str) -> Optional[Commit]:
        repo = self.__get_repo(path=path)
        if repo:
            try:
                return repo.commit(commit_hash)
            except BadName as ex:
                return None
        else:
            return None


    def _get_gitignore_path(self, path: str) -> Path:
        git_path = self.__file_system_service.get_path(path)
        return git_path.joinpath('.gitignore')


    def git_add_by_path(self, repo_path: str, file_path: str):
        repo = self.__get_repo(path=repo_path)
        if repo:
            abs_path = os.path.join(repo_path, file_path)
            if os.path.isfile(abs_path):
                repo.index.add([file_path])
            else:
                repo.index.remove([file_path])


    def git_add_all(self, repo_path: str):
        repo = self.__get_repo(path=repo_path)
        if repo:
            repo.git.add(all=True)


    def git_diff_stat(self, repo_path: str) -> str:
        repo = self.__get_repo(path=repo_path)
        if repo:
            try:
                diff: str = repo.git.diff("--cached", "--stat")
                if diff:
                    return diff.splitlines()[-1]
                else:
                    return None
            except ValueError as e:
                return str(e)


    def add_files_with_size_limit_or_warn(
        self,
        repo_path: str,
        files_to_add: Optional[str] = None,
        max_file_size: int = 500 * 1024,  # default 500KB
    ) -> bool:
        """
        Adds to git only files <= max_file_size.
        If files_to_add is not provided, adds all changed/new files except those that are too large.
        Returns True if files were added, False if operation was aborted due to large files.
        """
        if files_to_add:
            return self.add_files_by_path(repo_path=repo_path, files_to_add=files_to_add, max_file_size=max_file_size)

        return self.add_all_files(repo_path=repo_path, max_file_size=max_file_size)


    def add_files_by_path(
        self,
        repo_path: str,
        files_to_add: str,
        max_file_size: int = 500 * 1024,  # default 500KB
    ) -> bool:
        repo = self.__get_repo(path=repo_path)

        if files_to_add.strip().endswith(","):
            space_warning = f"[{ColorScheme.WARNING.value}][WARNING] Use only commas to separate files, without spaces[/{ColorScheme.WARNING.value}]"
            print(space_warning)
            return False
        
        files = [f.strip() for f in files_to_add.split(",") if f.strip()]

        rejected_file_paths = []
        missing_file_paths = []

        if repo.head.is_valid():
            staged_files = [item.a_path for item in repo.index.diff('HEAD')]
        else:
            staged_files = []

        deleted_files = [item.a_path for item in repo.index.diff(None) if item.change_type == 'D']

        for file_path in files:
            if file_path not in deleted_files and file_path not in staged_files:
                abs_path = os.path.join(repo_path, file_path)

                if not os.path.isfile(abs_path):
                    missing_file_paths.append(file_path)
                elif os.path.getsize(abs_path) > max_file_size:
                    rejected_file_paths.append(file_path)

        if missing_file_paths:
            not_found_file_warning = f"[{ColorScheme.WARNING.value}][WARNING] The following files do not exist and cannot be added: {', '.join(missing_file_paths)}[{ColorScheme.WARNING.value}]"
            print(not_found_file_warning)

        if rejected_file_paths:
            size_kb = max_file_size // 1024
            wrong_size_warning = f"[{ColorScheme.WARNING.value}][WARNING] The following files exceed {size_kb}KB and cannot be added: {', '.join(rejected_file_paths)}[{ColorScheme.WARNING.value}]"
            print(wrong_size_warning)

        if rejected_file_paths or missing_file_paths:
            return False

        for file_path in files:
            if file_path not in staged_files:
                self.git_add_by_path(repo_path=repo_path, file_path=file_path)

        return True


    def add_all_files(
        self,
        repo_path: str,
        max_file_size: int = 500 * 1024,  # default 500KB
    ) -> bool:
        repo = self.__get_repo(path=repo_path)

        files = [item.a_path for item in repo.index.diff(None)] + repo.untracked_files

        rejected_file_paths = []
        deleted_files = [item.a_path for item in repo.index.diff(None) if item.change_type == 'D']

        for file_path in files:
            abs_path = os.path.join(repo_path, file_path)
            if file_path not in deleted_files and os.path.getsize(abs_path) > max_file_size:
                rejected_file_paths.append(file_path)

        if rejected_file_paths:
            size_kb = max_file_size // 1024
            wrong_size_warning = f"[{ColorScheme.WARNING.value}][WARNING] The following files exceed {size_kb}KB and cannot be added: {', '.join(rejected_file_paths)}[{ColorScheme.WARNING.value}]"
            print(wrong_size_warning)
            return False

        self.git_add_all(repo_path=repo_path)

        return True


    def init_gitignore(self, path: str):
        gitignore_path = self._get_gitignore_path(path=path)
        if not gitignore_path.exists():
            self.__file_system_service.create_if_not_exists_file(gitignore_path)
            self.git_add_by_path(repo_path=path, file_path=str(gitignore_path))

        is_present_tsr = self.__file_system_service.find_line_in_text_file(file=str(gitignore_path),
                                                                           find=self.__git_ignore_thestage_line)
        if not is_present_tsr:
            self.__file_system_service.add_line_to_text_file(file=str(gitignore_path),
                                                             new_line=self.__git_ignore_thestage_line)


    def is_head_detached(self, path: str) -> bool:
        repo = self.__get_repo(path=path)
        if repo:
            return repo.head.is_detached


    def reset_hard(self, path: str, deploy_key_path: str, reset_to_origin: bool):
        repo = self.__get_repo(path=path)
        if repo:
            git_ssh_cmd = 'ssh -F /dev/null -o StrictHostKeyChecking=no -o IdentitiesOnly=yes -i %s' % deploy_key_path
            with repo.git.custom_environment(GIT_SSH_COMMAND=git_ssh_cmd):
                if reset_to_origin:
                    repo.git.reset('--hard', f'origin/{repo.active_branch.name}')
                    typer.echo(f'Branch "{repo.active_branch.name}" is now synced to its remote counterpart')
                else:
                    typer.echo('Simple branch reset is not implemented')


    # refers to a "headless commit" where something was committed while in detached head state and head is pointing at that commit
    def is_head_committed_in_headless_state(self, path: str) -> bool:
        repo = self.__get_repo(path=path)
        if repo:
            commit = repo.head.commit
            for branch in repo.heads:
                for commit_item in repo.iter_commits(branch):
                    if commit_item.hexsha == commit.hexsha:
                        return False
        return True


    def is_branch_exists(self, path: str, branch_name: str) -> bool:
        repo = self.__get_repo(path=path)
        if repo:
            for branch in repo.heads:
                if branch.name == branch_name:
                    return True

            for ref in repo.remotes.origin.refs:
                if ref.remote_head == branch_name:
                    typer.echo(f'Found remote branch "{branch_name}"')
                    return True
        return False


    def checkout_to_new_branch(self, path: str, branch_name: str):
        repo = self.__get_repo(path=path)
        if repo:
            repo.git.checkout("-b", branch_name)


    def is_commit_exists(self, repo, commit_sha) -> bool:
        try:
            repo.commit(commit_sha)
            return True
        except ValueError:
            return False


    def configure_local_ssh_command(self, path: str, deploy_key_path: str):
        repo = self.__get_repo(path=path)
        if repo:
            ssh_cmd = f'ssh -o StrictHostKeyChecking=accept-new -o IdentitiesOnly=yes -i {deploy_key_path}'
            with repo.config_writer() as writer:
                writer.set_value("core", "sshCommand", ssh_cmd)


from typing import Optional, List

from pydantic import Field, BaseModel, ConfigDict

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse


class ProjectDto(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    public_id: Optional[str] = Field(None, alias='publicId')
    slug: Optional[str] = Field(None, alias='slug')
    status: Optional[str] = Field(None, alias='status')
    git_repository_url: Optional[str] = Field(None, alias='gitRepositoryUrl')
    git_repository_name: Optional[str] = Field(None, alias='gitRepositoryName')
    last_commit_hash: Optional[str] = Field(None, alias='lastCommitHash')
    last_commit_description: Optional[str] = Field(None, alias='lastCommitDescription')


class ProjectViewResponse(TheStageBaseResponse):
    project: Optional[ProjectDto] = Field(None, alias='project')
    isDefaultContainerDeletedPermanently: Optional[bool] = Field(None, alias='isDefaultContainerDeletedPermanently')



from enum import Enum
from typing import Optional, List

from pydantic import BaseModel, Field


class CliCommand(str, Enum):
    VERSION = "VERSION"
    CONNECT = "CONNECT"
    PROJECT_CLONE = "PROJECT_CLONE"
    PROJECT_INIT = "PROJECT_INIT"
    PROJECT_RUN = "PROJECT_RUN"
    PROJECT_CHECKOUT = "PROJECT_CHECKOUT"
    PROJECT_PULL = "PROJECT_PULL"
    PROJECT_RESET = "PROJECT_RESET"
    PROJECT_INFERENCE_SIMULATOR_RUN = "PROJECT_INFERENCE_SIMULATOR_RUN"
    PROJECT_INFERENCE_SIMULATOR_SAVE_METADATA = "PROJECT_INFERENCE_SIMULATOR_SAVE_METADATA"
    PROJECT_INFERENCE_SIMULATOR_PUSH = "PROJECT_INFERENCE_SIMULATOR_PUSH"
    PROJECT_INFERENCE_SIMULATOR_LS = "PROJECT_INFERENCE_SIMULATOR_LS"
    PROJECT_INFERENCE_SIMULATOR_LOGS = "PROJECT_INFERENCE_SIMULATOR_LOGS"
    PROJECT_MODEL_LS = "PROJECT_MODEL_LS"
    PROJECT_MODEL_DEPLOY_INSTANCE = "PROJECT_MODEL_DEPLOY_INSTANCE"
    PROJECT_MODEL_DEPLOY_SAGEMAKER = "PROJECT_MODEL_DEPLOY_SAGEMAKER"
    PROJECT_TASK_CANCEL = "PROJECT_TASK_CANCEL"
    PROJECT_TASK_LS = "PROJECT_TASK_LS"
    PROJECT_TASK_LOGS = "PROJECT_TASK_LOGS"
    PROJECT_CONFIG_SET_DEFAULT_CONTAINER = "PROJECT_CONFIG_SET_DEFAULT_CONTAINER"
    PROJECT_CONFIG_GET = "PROJECT_CONFIG_GET"
    PROJECT_ATTACH_INSTANCE = "PROJECT_ATTACH_INSTANCE"
    PROJECT_LS = "PROJECT_LS"
    PROJECT_DETACH_INSTANCE = "PROJECT_DETACH_INSTANCE"
    PROJECT_DELETE = "PROJECT_DELETE"
    CONTAINER_LS = "CONTAINER_LS"
    CONTAINER_INFO = "CONTAINER_INFO"
    CONTAINER_CONNECT = "CONTAINER_CONNECT"
    CONTAINER_UPLOAD = "CONTAINER_UPLOAD"
    CONTAINER_DOWNLOAD = "CONTAINER_DOWNLOAD"
    CONTAINER_CREATE = "CONTAINER_CREATE"
    CONTAINER_START = "CONTAINER_START"
    CONTAINER_STOP = "CONTAINER_STOP"
    CONTAINER_DELETE = "CONTAINER_DELETE"
    CONTAINER_RESTART = "CONTAINER_RESTART"
    CONTAINER_LOGS = "CONTAINER_LOGS"
    CONTAINER_IMAGE_LS = "CONTAINER_IMAGE_LS"
    INSTANCE_RENTED_LS = "INSTANCE_RENTED_LS"
    INSTANCE_RENTED_CONNECT = "INSTANCE_RENTED_CONNECT"
    INSTANCE_RENTED_CREATE = "INSTANCE_RENTED_CREATE"
    INSTANCE_RENTED_TERMINATE = "INSTANCE_RENTED_TERMINATE"
    INSTANCE_RENTED_RESTART = "INSTANCE_RENTED_RESTART"
    INSTANCE_SELF_HOSTED_LS = "INSTANCE_SELF_HOSTED_LS"
    INSTANCE_SELF_HOSTED_CONNECT = "INSTANCE_SELF_HOSTED_CONNECT"
    CONFIG_GET = "CONFIG_GET"
    CONFIG_SET = "CONFIG_SET"
    CONFIG_CLEAR = "CONFIG_CLEAR"
    CONFIG_UPLOAD_SSH_KEY = "CONFIG_UPLOAD_SSH_KEY"


class CliCommandAvailability(str, Enum):
    ALLOWED = "ALLOWED"
    RESTRICTED = "RESTRICTED"
    DEPRECATED = "DEPRECATED"


ALWAYS_AVAILABLE_COMMANDS = [
    CliCommand.VERSION,
    CliCommand.PROJECT_CONFIG_GET,
    CliCommand.CONFIG_GET,
    CliCommand.CONFIG_SET,
    CliCommand.CONFIG_CLEAR,
]


class CliCommandMetadata(BaseModel):
    instanceOfferProviderNames: Optional[List[str]] = Field(None, alias='instanceOfferProviderNames')
    instanceOfferGpuNames: Optional[List[str]] = Field(None, alias='instanceOfferGpuNames')



from typing import Optional

from thestage.project.dto.project_entity import ProjectEntity
from thestage.project.dto.project_response import ProjectDto
from thestage.services.abstract_mapper import AbstractMapper

class ProjectMapper(AbstractMapper):

    def build_entity(self, item: ProjectDto) -> Optional[ProjectEntity]:
        if not item:
            return None

        return ProjectEntity(
            public_id=item.public_id or '',
            slug=item.slug or '',
            status=item.status or '',
            last_commit_hash=item.last_commit_hash or ''
        )


from typing import Optional

from thestage.inference_model.dto.inference_model_entity import InferenceModelEntity
from thestage.services.abstract_mapper import AbstractMapper
from thestage.inference_model.dto.inference_model import InferenceModel


class InferenceModelMapper(AbstractMapper):
    def build_entity(self, item: InferenceModel) -> Optional[InferenceModelEntity]:
        if not item:
            return None

        return InferenceModelEntity(
            public_id=item.public_id,
            slug=item.slug,
            status=item.status or '',
            commit_hash=item.commit_hash or '',
            environment_metadata=item.environment_metadata or {},
            started_at=item.created_at or '',
            finished_at=item.updated_at or ''  # TODO updated_at cannot be finished_at
        )



from typing import Optional

from thestage.inference_simulator.dto.inference_simulator_entity import InferenceSimulatorEntity
from thestage.services.abstract_mapper import AbstractMapper
from thestage.inference_simulator.dto.inference_simulator import InferenceSimulator


class InferenceSimulatorMapper(AbstractMapper):

    def build_entity(self, item: InferenceSimulator) -> Optional[InferenceSimulatorEntity]:
        if not item:
            return None

        return InferenceSimulatorEntity(
            public_id=item.public_id or '',
            slug=item.slug or '',
            status=item.status or '',
            http_endpoint=item.http_endpoint or '',
            grpc_endpoint=item.grpc_endpoint or '',
            metrics_endpoint=item.metrics_endpoint or '',
            started_at=item.created_at or '',
        )



from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class DockerImageEntity(BaseModel):

    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
    )

    image: Optional[str] = Field(None, alias='IMAGE')
    tag: Optional[str] = Field(None, alias='TAG')



from typing import List, Optional

from pydantic import Field, ConfigDict, BaseModel


class ModifyProjectInstanceAssignmentRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    projectPublicId: Optional[str] = Field(None, alias='projectPublicId')
    projectSlug: Optional[str] = Field(None, alias='projectSlug')

    instanceRentedPublicIdsToAssign: Optional[List[str]] = Field(None, alias='instanceRentedPublicIdsToAssign')
    instanceRentedSlugsToAssign: Optional[List[str]] = Field(None, alias='instanceRentedSlugsToAssign')
    instanceRentedPublicIdsToDelete: Optional[List[str]] = Field(None, alias='instanceRentedPublicIdsToDelete')
    instanceRentedSlugsToDelete: Optional[List[str]] = Field(None, alias='instanceRentedSlugsToDelete')

    selfhostedInstancePublicIdsToAssign: Optional[List[str]] = Field(None, alias='selfhostedInstancePublicIdsToAssign')
    selfhostedInstanceSlugsToAssign: Optional[List[str]] = Field(None, alias='selfhostedInstanceSlugsToAssign')
    selfhostedInstancePublicIdsToDelete: Optional[List[str]] = Field(None, alias='selfhostedInstancePublicIdsToDelete')
    selfhostedInstanceSlugsToDelete: Optional[List[str]] = Field(None, alias='selfhostedInstanceSlugsToDelete')